from platypush.plugins import action
from platypush.plugins.http.request import HttpRequestPlugin


class WeatherDarkskyPlugin(HttpRequestPlugin):
    """
    Plugin for getting weather updates through Darksky API

    Requires:

        * **requests** (``pip install requests``)
    """

    def __init__(self, darksky_token, lat, long, units='si', **kwargs):
        """
        :param darksky_token: Your token for using the darksky API, see https://darksky.net/dev
        :type darksky_token: str

        :param lat: Default forecast latitude
        :type lat: float

        :param long: Default forecast longitude
        :type long: float

        :param units: Weather units (default: "si").

        Supported units:

            * **si** (international system)
            * **us** (US imperial units)
            * **uk** (UK imperial units)
            * **ca** (Canada imperial units)

        :type units: str
        """

        super().__init__(method='get', output='json')
        self.darksky_token = darksky_token
        self.units = units
        self.lat = lat
        self.long = long
        self.latest_bulletin = {}

    def _get_url(self, lat=None, long=None):
        return 'https://api.darksky.net/forecast/{}/{},{}?units={}'. \
            format(self.darksky_token, (lat or self.lat), (long or self.long),
                   self.units)

    @action
    def get_current_weather(self, lat=None, long=None, **kwargs):
        """
        Get the current weather.

        :param lat: Weather latitude (default: configured latitude)
        :type lat: float

        :param long: Weather longitude (default: configured longitude)
        :type long: float

        :returns: A dictionary containing the current weather object.

        Example output::

                output = {
                    "time": 1529947892,
                    "summary": "Mostly Cloudy",
                    "icon": "partly-cloudy-day",
                    "precipIntensity": 0.0483,
                    "precipProbability": 0.04,
                    "precipType": "rain",
                    "temperature": 27.94,
                    "apparentTemperature": 29.6,
                    "dewPoint": 20.01,
                    "humidity": 0.62,
                    "pressure": 1009.34,
                    "windSpeed": 1.83,
                    "windGust": 5.49,
                    "windBearing": 192,
                    "cloudCover": 0.66,
                    "uvIndex": 0,
                    "visibility": 16.09,
                    "ozone": 273.74
                }
        """

        response = self.get(self._get_url(lat, long))
        return response.output['currently']

    @action
    def get_hourly_forecast(self, lat=None, long=None, **kwargs):
        """
        Get the hourly forecast.

        :param lat: Weather latitude (default: configured latitude)
        :type lat: float

        :param long: Weather longitude (default: configured longitude)
        :type long: float

        :returns: A forecast object.

        Example output::

            output = {
                "summary": "Partly cloudy starting tomorrow morning, continuing until tomorrow evening.",
                "icon": "partly-cloudy-day",
                "data": [
                    {
                        "time": 1529946000,
                        "summary": "Clear",
                        "icon": "clear-day",
                        "precipIntensity": 0,
                        "precipProbability": 0,
                        "temperature": 18.94,
                        "apparentTemperature": 18.94,
                        "dewPoint": 11.99,
                        "humidity": 0.64,
                        "pressure": 1025.53,
                        "windSpeed": 5.1,
                        "windGust": 6.22,
                        "windBearing": 329,
                        "cloudCover": 0.14,
                        "uvIndex": 1,
                        "visibility": 14.19,
                        "ozone": 334.3
                    },
                    {
                        "time": 1529949600,
                        "summary": "Clear",
                        "icon": "clear-day",
                        "precipIntensity": 0,
                        "precipProbability": 0,
                        "temperature": 18.41,
                        "apparentTemperature": 18.41,
                        "dewPoint": 11.12,
                        "humidity": 0.63,
                        "pressure": 1025.54,
                        "windSpeed": 4.6,
                        "windGust": 6.18,
                        "windBearing": 340,
                        "cloudCover": 0.07,
                        "uvIndex": 1,
                        "visibility": 16.09,
                        "ozone": 333.53
                    },
                    # ...
            }
        """

        response = self.get(self._get_url(lat, long))
        return response.output['hourly']

    @action
    def get_daily_forecast(self, lat=None, long=None, **kwargs):
        """
        Get the daily forecast.

        :param lat: Weather latitude (default: configured latitude)
        :type lat: float

        :param long: Weather longitude (default: configured longitude)
        :type long: float

        :returns: A forecast object.

        Example output::

            "output": {
                "summary": "Light rain on Sunday, with high temperatures rising to 28°C on Sunday.",
                "icon": "rain",
                "data": [
                    {
                        "time": 1529877600,
                        "summary": "Mostly cloudy until afternoon.",
                        "icon": "partly-cloudy-day",
                        "sunriseTime": 1529896835,
                        "sunsetTime": 1529957280,
                        "moonPhase": 0.42,
                        "precipIntensity": 0,
                        "precipIntensityMax": 0.0051,
                        "precipIntensityMaxTime": 1529888400,
                        "precipProbability": 0,
                        "temperatureHigh": 20.04,
                        "temperatureHighTime": 1529931600,
                        "temperatureLow": 10.68,
                        "temperatureLowTime": 1529982000,
                        "apparentTemperatureHigh": 20.04,
                        "apparentTemperatureHighTime": 1529931600,
                        "apparentTemperatureLow": 10.68,
                        "apparentTemperatureLowTime": 1529982000,
                        "dewPoint": 12.18,
                        "humidity": 0.77,
                        "pressure": 1025.16,
                        "windSpeed": 3.84,
                        "windGust": 6.51,
                        "windGustTime": 1529881200,
                        "windBearing": 336,
                        "cloudCover": 0.5,
                        "uvIndex": 6,
                        "uvIndexTime": 1529928000,
                        "visibility": 14.08,
                        "ozone": 331.24,
                        "temperatureMin": 13.89,
                        "temperatureMinTime": 1529960400,
                        "temperatureMax": 20.04,
                        "temperatureMaxTime": 1529931600,
                        "apparentTemperatureMin": 13.89,
                        "apparentTemperatureMinTime": 1529960400,
                        "apparentTemperatureMax": 20.04,
                        "apparentTemperatureMaxTime": 1529931600
                    },
                    {
                        "time": 1529964000,
                        "summary": "Partly cloudy throughout the day.",
                        "icon": "partly-cloudy-day",
                        "sunriseTime": 1529983261,
                        "sunsetTime": 1530043677,
                        "moonPhase": 0.45,
                        "precipIntensity": 0,
                        "precipIntensityMax": 0,
                        "precipProbability": 0,
                        "temperatureHigh": 20.95,
                        "temperatureHighTime": 1530018000,
                        "temperatureLow": 11.47,
                        "temperatureLowTime": 1530064800,
                        "apparentTemperatureHigh": 20.95,
                        "apparentTemperatureHighTime": 1530018000,
                        "apparentTemperatureLow": 11.47,
                        "apparentTemperatureLowTime": 1530064800,
                        "dewPoint": 10.19,
                        "humidity": 0.69,
                        "pressure": 1026.14,
                        "windSpeed": 3.67,
                        "windGust": 7.13,
                        "windGustTime": 1530036000,
                        "windBearing": 4,
                        "cloudCover": 0.3,
                        "uvIndex": 5,
                        "uvIndexTime": 1530010800,
                        "visibility": 16.09,
                        "ozone": 328.59,
                        "temperatureMin": 10.68,
                        "temperatureMinTime": 1529982000,
                        "temperatureMax": 20.95,
                        "temperatureMaxTime": 1530018000,
                        "apparentTemperatureMin": 10.68,
                        "apparentTemperatureMinTime": 1529982000,
                        "apparentTemperatureMax": 20.95,
                        "apparentTemperatureMaxTime": 1530018000
                    },
                    # ...
                }
        """

        response = self.get(self._get_url(lat, long))
        return response.output['daily']


# vim:sw=4:ts=4:et:




import json
import threading
import time

from platypush.backend.http.utils import HttpUtils
from platypush.plugins import Plugin, action
from platypush.procedure import Procedure


class UtilsPlugin(Plugin):
    """
    A plugin for general-purpose util methods
    """

    _DEFAULT_TIMEOUT_PREFIX = '_PlatypushTimeout_'
    _timeout_hndl_idx = 0
    _timeout_hndl_idx_lock = threading.RLock()

    _DEFAULT_INTERVAL_PREFIX = '_PlatypushInterval_'
    _interval_hndl_idx = 0
    _interval_hndl_idx_lock = threading.RLock()

    _pending_timeouts = {}
    _pending_intervals = {}
    _pending_timeouts_lock = threading.RLock()
    _pending_intervals_lock = threading.RLock()

    @action
    def sleep(self, seconds):
        """
        Make the current executor sleep for the specified number of seconds.

        :param seconds: Sleep seconds
        :type seconds: float
        """

        time.sleep(seconds)

    @action
    def set_timeout(self, seconds, actions, name=None, **args):
        """
        Define a set of actions to run after the specified amount of `seconds`.

        :param seconds: Number of seconds before running the timeout procedure
        :type seconds: float

        :param actions: List of actions to be executed after the timeout expires
        :type actions: list[dict]

        :param name: Set an optional name for this timeout. It is advised to set
            a name if you are planning to programmatically cancel the timeout in
            your business logic.
        :type name: str

        :param args: Optional arguments/context to pass to the timeout function
        """

        with self._timeout_hndl_idx_lock:
            self._timeout_hndl_idx += 1
            if not name:
                name = self._DEFAULT_TIMEOUT_PREFIX + str(self._timeout_hndl_idx)
            if name in self._pending_timeouts:
                return (None,
                        "A timeout named '{}' is already awaiting".format(name))

        procedure = Procedure.build(name=name, requests=actions, _async=False)
        self._pending_timeouts[name] = procedure

        def _proc_wrapper(procedure, **kwargs):
            try:
                procedure.execute(**kwargs)
            finally:
                with self._pending_timeouts_lock:
                    if name in self._pending_timeouts:
                        del self._pending_timeouts[name]

        with self._pending_timeouts_lock:
            self._pending_timeouts[name] = threading.Timer(seconds,
                                                           _proc_wrapper,
                                                           args=[procedure],
                                                           kwargs=args)
        self._pending_timeouts[name].start()

    @action
    def clear_timeout(self, name):
        """
        Clear a pending timeout procedure

        :param name: Name of the timeout to clear
        :type name: str
        """

        timer = None

        with self._pending_timeouts_lock:
            if name not in self._pending_timeouts:
                self.logger.debug('{} is not a pending timeout'.format(name))
                return
            timer = self._pending_timeouts.pop(name)

        timer.cancel()

    @action
    def get_timeouts(self):
        """
        Get info about the pending timeouts

        :returns: dict.

        Example::

            {
                "test_timeout": {
                    "seconds": 10.0,
                    "actions": [
                        {
                            "action": "action_1",
                            "args": {
                                "name_1": "value_1"
                            }
                        }
                    ]
                }
            }

        """

        response = {}

        for name in self._pending_timeouts.keys():
            response[name] = self.get_timeout(name).output.get(name)
        return response

    @action
    def get_timeout(self, name):
        """
        Get info about a pending timeout

        :param name: Name of the timeout to get
        :type name: str

        :returns: dict

        Example::

            {
                "test_timeout": {
                    "seconds": 10.0,
                    "actions": [
                        {
                            "action": "action_1",
                            "args": {
                                "name_1": "value_1"
                            }
                        }
                    ]
                }
            }

        If no such timeout exist with the specified name then the value of the
        timeout name will be null.
        """

        response = { name: None }

        with self._pending_timeouts_lock:
            timer = self._pending_timeouts.get(name)
            if not timer:
                return response

            return {
                name: {
                    'seconds': timer.interval,
                    'actions': [
                        json.loads(str(a)) for a in timer.args[0].requests
                    ]
                }
            }

    @action
    def set_interval(self, seconds, actions, name=None, **args):
        """
        Define a set of actions to run each specified amount of `seconds`.

        :param seconds: Number of seconds between two runs of the interval
            procedure
        :type seconds: float

        :param actions: List of actions to be executed at each interval
        :type actions: list[dict]

        :param name: Set an optional name for this interval. It is advised to
            set a name if you are planning to programmatically cancel the
            interval in your business logic.
        :type name: str

        :param args: Optional arguments/context to pass to the interval function
        """

        with self._interval_hndl_idx_lock:
            self._interval_hndl_idx += 1
            if not name:
                name = self._DEFAULT_INTERVAL_PREFIX + \
                    str(self._interval_hndl_idx)

            if name in self._pending_intervals:
                return (None,
                        "An interval named '{}' is already running".format(name))


        procedure = Procedure.build(name=name, requests=actions, _async=False)
        self._pending_intervals[name] = procedure

        def _proc_wrapper(procedure, seconds, **kwargs):
            while True:
                with self._pending_intervals_lock:
                    if name not in self._pending_intervals:
                        return

                procedure.execute(**kwargs)
                time.sleep(seconds)

        with self._pending_intervals_lock:
            self._pending_intervals[name] = threading.Thread(
                target=_proc_wrapper, args=[procedure, seconds], kwargs=args)
        self._pending_intervals[name].start()

    @action
    def clear_interval(self, name):
        """
        Clear a running interval procedure

        :param name: Name of the interval to clear
        :type name: str
        """

        interval = None

        with self._pending_intervals_lock:
            if name not in self._pending_intervals:
                self.logger.debug('{} is not a running interval'.format(name))
                return
            del self._pending_intervals[name]


    @action
    def get_intervals(self):
        """
        Get info about the running intervals

        :returns: dict

        Example::

            {
                "test_interval": {
                    "seconds": 10.0,
                    "actions": [
                        {
                            "action": "action_1",
                            "args": {
                                "name_1": "value_1"
                            }
                        }
                    ]
                }
            }

        """

        response = {}

        for name in self._pending_intervals.keys():
            response[name] = self.get_interval(name).output.get(name)
        return response

    @action
    def get_interval(self, name):
        """
        Get info about a running interval

        :param name: Name of the interval to get
        :type name: str

        :returns: dict. Example:

        .. code-block:: json

            {
                "test_interval": {
                    "seconds": 10.0,
                    "actions": [
                        {
                            "action": "action_1",
                            "args": {
                                "name_1": "value_1"
                            }
                        }
                    ]
                }
            }

        If no such interval exist with the specified name then the value of the
        timeout name will be null.
        """

        response = { name: None }

        with self._pending_intervals_lock:
            timer = self._pending_intervals.get(name)
            if not timer:
                return response

            return {
                name: {
                    'seconds': timer._args[1],
                    'actions': [
                        json.loads(str(a)) for a in timer._args[0].requests
                    ]
                }
            }

    @action
    def search_directory(self, directory, extensions, recursive=False):
        return HttpUtils.search_directory(directory, recursive=recursive, *extensions)

    @action
    def search_web_directory(self, directory, extensions):
        return HttpUtils.search_web_directory(directory, *extensions)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event
from platypush.message.event.media import MediaEvent


class VideoEvent(MediaEvent):
    """ Base class for video events """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VideoPlayEvent(VideoEvent):
    """
    Event triggered when a new video content is played
    """

    def __init__(self, video=None, *args, **kwargs):
        """
        :param video: File name or URI of the played video
        :type video: str
        """

        super().__init__(*args, video=video, **kwargs)


class VideoStopEvent(VideoEvent):
    """
    Event triggered when a video is stopped
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VideoPauseEvent(VideoEvent):
    """
    Event triggered when a video playback is paused
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NewPlayingVideoEvent(VideoEvent):
    """
    Event triggered when a video playback is paused
    """

    def __init__(self, video=None, *args, **kwargs):
        """
        :param video: File name or URI of the played video
        :type video: str
        """

        super().__init__(*args, video=video, **kwargs)


# vim:sw=4:ts=4:et:



#!/usr/bin/env python
"""
LIS3DH Python Library for Raspberry Pi
Created by Matt Dyson (mattdyson.org)
Project hosted at https://github.com/mattdy/python-lis3dh
Version 1.0 - 10/01/16
Version 1.1 - 19/03/16 (Mal Smalley) Adding click detection
Version 1.2 - 21/01/18 (holzfigure) Using new Adafruit library Adafruit_GPIO;
                                    Changing default busnumber (-1 -> 1);
                                    Fixing indentation issues;
                                    Increasing PEP8 and Python3 compatibility;
                                    Replacing variable name "range"
                                      (which is a keyword) with "g-range";

Requires the Adafruit_Python_GPIO library
https://github.com/adafruit/Adafruit_Python_GPIO

Inspiration and assistance from:
 - https://github.com/adafruit/Adafruit_LIS3DH
 - https://www.adafruit.com/datasheets/LIS3DH.pdf
"""

from Adafruit_GPIO import I2C
import RPi.GPIO as GPIO  # needed for Hardware interrupt


class LIS3DH:

    # I2C
    i2c = None
    I2C_ADDRESS_1 = 0x18
    I2C_ADDRESS_2 = 0x19
    # default
    I2C_DEFAULT = I2C_ADDRESS_1

    # bus
    BUS_NUMBER = 1  # -1

    # Ranges
    RANGE_2G  = 0b00  # default
    RANGE_4G  = 0b01
    RANGE_8G  = 0b10
    RANGE_16G = 0b11
    # default
    RANGE_DEFAULT = RANGE_2G

    # Refresh rates
    DATARATE_400HZ          = 0b0111  # 400Hz  # default
    DATARATE_200HZ          = 0b0110  # 200Hz
    DATARATE_100HZ          = 0b0101  # 100Hz
    DATARATE_50HZ           = 0b0100  # 50Hz
    DATARATE_25HZ           = 0b0011  # 25Hz
    DATARATE_10HZ           = 0b0010  # 10Hz
    DATARATE_1HZ            = 0b0001  # 1Hz
    DATARATE_POWERDOWN      = 0       # Power down
    DATARATE_LOWPOWER_1K6HZ = 0b1000  # Low power mode (1.6KHz)
    DATARATE_LOWPOWER_5KHZ  = 0b1001  # Low power mode (5KHz) / Normal power mode (1.25KHz)
    # default
    DATARATE_DEFAULT = DATARATE_400HZ

    # Registers
    REG_STATUS1       = 0x07
    REG_OUTADC1_L     = 0x08
    REG_OUTADC1_H     = 0x09
    REG_OUTADC2_L     = 0x0A
    REG_OUTADC2_H     = 0x0B
    REG_OUTADC3_L     = 0x0C
    REG_OUTADC3_H     = 0x0D
    REG_INTCOUNT      = 0x0E
    REG_WHOAMI        = 0x0F  # Device identification register
    REG_TEMPCFG       = 0x1F
    REG_CTRL1         = 0x20  # Used for data rate selection, and enabling/disabling individual axis
    REG_CTRL2         = 0x21
    REG_CTRL3         = 0x22
    REG_CTRL4         = 0x23  # Used for BDU, scale selection, resolution selection and self-testing
    REG_CTRL5         = 0x24
    REG_CTRL6         = 0x25
    REG_REFERENCE     = 0x26
    REG_STATUS2       = 0x27
    REG_OUT_X_L       = 0x28
    REG_OUT_X_H       = 0x29
    REG_OUT_Y_L       = 0x2A
    REG_OUT_Y_H       = 0x2B
    REG_OUT_Z_L       = 0x2C
    REG_OUT_Z_H       = 0x2D
    REG_FIFOCTRL      = 0x2E
    REG_FIFOSRC       = 0x2F
    REG_INT1CFG       = 0x30
    REG_INT1SRC       = 0x31
    REG_INT1THS       = 0x32
    REG_INT1DUR       = 0x33
    REG_CLICKCFG      = 0x38
    REG_CLICKSRC      = 0x39
    REG_CLICKTHS      = 0x3A
    REG_TIMELIMIT     = 0x3B
    REG_TIMELATENCY   = 0x3C
    REG_TIMEWINDOW    = 0x3D

    # Values
    DEVICE_ID  = 0x33
    INT_IO     = 0x04  # GPIO pin for interrupt
    CLK_NONE   = 0x00
    CLK_SINGLE = 0x01
    CLK_DOUBLE = 0x02

    AXIS_X = 0x00
    AXIS_Y = 0x01
    AXIS_Z = 0x02

    # changed busnumber to 1 (from -1)
    # alternative i2c address=0x19
    def __init__(self, address=I2C_DEFAULT, bus=BUS_NUMBER,
                 g_range=RANGE_DEFAULT, datarate=DATARATE_DEFAULT,
                 debug=False):
        self.isDebug = debug
        self.debug("Initialising LIS3DH")

        # self.i2c = Adafruit_I2C(address, busnum=bus)
        self.i2c = I2C.Device(address, busnum=bus)
        self.address = address

        try:
            val = self.i2c.readU8(self.REG_WHOAMI)
            if val != self.DEVICE_ID:
                # raise Exception(("Device ID incorrect - expected 0x%X, " +
                #                  "got 0x%X at address 0x%X") % (
                #                          self.DEVICE_ID, val, self.address))
                raise Exception(("Device ID incorrect - expected 0x{:x}, " +
                                 "got 0x{:x} at address 0x{:x}").format(
                                     self.DEVICE_ID, val, self.address))

            self.debug(("Successfully connected to LIS3DH " +
                        "at address 0x{:x}").format(self.address))
        except Exception as e:
            print("Error establishing connection with LIS3DH")
            print(e)

        # Enable all axis
        self.setAxisStatus(self.AXIS_X, True)
        self.setAxisStatus(self.AXIS_Y, True)
        self.setAxisStatus(self.AXIS_Z, True)

        # Set refresh rate (default: 400Hz)
        self.setDataRate(datarate)

        self.setHighResolution()
        self.setBDU()

        self.setRange(g_range)

    # Get reading from X axis
    def getX(self):
        return self.getAxis(self.AXIS_X)

    # Get reading from Y axis
    def getY(self):
        return self.getAxis(self.AXIS_Y)

    # Get reading from Z axis
    def getZ(self):
        return self.getAxis(self.AXIS_Z)

    # Get a reading from the desired axis
    def getAxis(self, axis):
        # Determine which register we need to read from (2 per axis)
        base = self.REG_OUT_X_L + (2 * axis)

        # Read the first register (lower bits)
        low = self.i2c.readU8(base)
        # Read the next register (higher bits)
        high = self.i2c.readU8(base + 1)
        # Combine the two components
        res = low | (high << 8)
        # Calculate the twos compliment of the result
        res = self.twosComp(res)

        # Fetch the range we're set to, so we can
        # accurately calculate the result
        g_range = self.getRange()
        divisor = 1
        if g_range == self.RANGE_2G:    divisor = 16380
        elif g_range == self.RANGE_4G:  divisor = 8190
        elif g_range == self.RANGE_8G:  divisor = 4096
        elif g_range == self.RANGE_16G: divisor = 1365.33

        return float(res) / divisor

    # Get the range that the sensor is currently set to
    def getRange(self):
        val = self.i2c.readU8(self.REG_CTRL4)  # Get value from register
        val = (val >> 4)  # Remove lowest 4 bits
        val &= 0b0011  # Mask off two highest bits

        if val == self.RANGE_2G:   return self.RANGE_2G
        elif val == self.RANGE_4G: return self.RANGE_4G
        elif val == self.RANGE_8G: return self.RANGE_8G
        else:                      return self.RANGE_16G

    # Set the range of the sensor (2G, 4G, 8G, 16G)
    def setRange(self, g_range):
        if g_range < 0 or g_range > 3:
            raise Exception("Tried to set invalid range")

        val = self.i2c.readU8(self.REG_CTRL4)  # Get value from register
        val &= ~(0b110000)  # Mask off lowest 4 bits
        val |= (g_range << 4)  # Write in our new range
        self.writeRegister(self.REG_CTRL4, val)  # Write back to register

    # Enable or disable an individual axis
    # Read status from CTRL_REG1, then write back with
    # appropriate status bit changed
    def setAxisStatus(self, axis, enable):
        if axis < 0 or axis > 2:
            raise Exception("Tried to modify invalid axis")

        current = self.i2c.readU8(self.REG_CTRL1)
        status = 1 if enable else 0
        final = self.setBit(current, axis, status)
        self.writeRegister(self.REG_CTRL1, final)

    def setInterrupt(self, mycallback):
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self.INT_IO, GPIO.IN)
        GPIO.add_event_detect(self.INT_IO, GPIO.RISING, callback=mycallback)

    def setClick(self, clickmode, clickthresh=80,
                 timelimit=10, timelatency=20, timewindow=100,
                 mycallback=None):
        if (clickmode == self.CLK_NONE):
            val = self.i2c.readU8(self.REG_CTRL3)  # Get value from register
            val &= ~(0x80)  # unset bit 8 to disable interrupt
            self.writeRegister(self.REG_CTRL3, val)  # Write back to register
            self.writeRegister(self.REG_CLICKCFG, 0)  # disable all interrupts
            return
        self.writeRegister(self.REG_CTRL3, 0x80)  # turn on int1 click
        self.writeRegister(self.REG_CTRL5, 0x08)  # latch interrupt on int1

        if (clickmode == self.CLK_SINGLE):
            # turn on all axes & singletap
            self.writeRegister(self.REG_CLICKCFG, 0x15)
        if (clickmode == self.CLK_DOUBLE):
            # turn on all axes & doubletap
            self.writeRegister(self.REG_CLICKCFG, 0x2A)

        # set timing parameters
        self.writeRegister(self.REG_CLICKTHS, clickthresh)
        self.writeRegister(self.REG_TIMELIMIT, timelimit)
        self.writeRegister(self.REG_TIMELATENCY, timelatency)
        self.writeRegister(self.REG_TIMEWINDOW, timewindow)

        if mycallback is not None:
            self.setInterrupt(mycallback)

    def getClick(self):
        reg = self.i2c.readU8(self.REG_CLICKSRC)  # read click register
        self.i2c.readU8(self.REG_INT1SRC)         # reset interrupt flag
        return reg

    # Set the rate (cycles per second) at which data is gathered
    def setDataRate(self, dataRate):
        val = self.i2c.readU8(self.REG_CTRL1)  # Get current value
        val &= 0b1111  # Mask off lowest 4 bits
        val |= (dataRate << 4)  # Write in our new data rate to highest 4 bits
        self.writeRegister(self.REG_CTRL1, val)  # Write back to register

    # Set whether we want to use high resolution or not
    def setHighResolution(self, highRes=True):
        val = self.i2c.readU8(self.REG_CTRL4)  # Get current value
        status = 1 if highRes else 0

        # High resolution is bit 4 of REG_CTRL4
        final = self.setBit(val, 3, status)
        self.writeRegister(self.REG_CTRL4, final)

    # Set whether we want to use block data update or not
    # False = output registers not updated until MSB and LSB reading
    def setBDU(self, bdu=True):
        val = self.i2c.readU8(self.REG_CTRL4)  # Get current value
        status = 1 if bdu else 0

        # Block data update is bit 8 of REG_CTRL4
        final = self.setBit(val, 7, status)
        self.writeRegister(self.REG_CTRL4, final)

    # Write the given value to the given register
    def writeRegister(self, register, value):
        self.debug("WRT {} to register 0x{:x}".format(
            bin(value), register))
        self.i2c.write8(register, value)

    # Set the bit at index 'bit' to 'value' on 'input' and return
    def setBit(self, input, bit, value):
        mask = 1 << bit
        input &= ~mask
        if value:
            input |= mask
        return input

    # Return a 16-bit signed number (two's compliment)
    # Thanks to http://stackoverflow.com/questions/16124059/trying-to-
    #   read-a-twos-complement-16bit-into-a-signed-decimal
    def twosComp(self, x):
        if (0x8000 & x):
            x = - (0x010000 - x)
        return x

    # Print an output of all registers
    def dumpRegisters(self):
        for x in range(0x0, 0x3D):
            read = self.i2c.readU8(x)
            print("{:x}: {}".format(x, bin(read)))

    def debug(self, message):
        if not self.isDebug:
            return
        print(message)



import json
import socket

from enum import Enum
from threading import Thread

from platypush.backend import Backend
from platypush.context import get_backend


class CameraPiBackend(Backend):
    """
    Backend to interact with a Raspberry Pi camera. It can start and stop
    recordings and take pictures. It can be programmatically controlled through
    the :class:`platypush.plugins.camera.pi` plugin. Note that the Redis backend
    must be configured and running to enable camera control.

    Requires:

        * **picamera** (``pip install picamera``)
        * **redis** (``pip install redis``) for inter-process communication with the camera process
    """

    class CameraAction(Enum):
        START_RECORDING = 'START_RECORDING'
        STOP_RECORDING = 'STOP_RECORDING'
        TAKE_PICTURE = 'TAKE_PICTURE'

        def __eq__(self, other):
            return self.value == other

    # noinspection PyUnresolvedReferences,PyPackageRequirements
    def __init__(self, listen_port, x_resolution=640, y_resolution=480,
                 redis_queue='platypush/camera/pi',
                 start_recording_on_startup=True,
                 framerate=24, hflip=False, vflip=False,
                 sharpness=0, contrast=0, brightness=50,
                 video_stabilization=False, iso=0, exposure_compensation=0,
                 exposure_mode='auto', meter_mode='average', awb_mode='auto',
                 image_effect='none', color_effects=None, rotation=0,
                 crop=(0.0, 0.0, 1.0, 1.0), **kwargs):
        """
        See https://www.raspberrypi.org/documentation/usage/camera/python/README.md
        for a detailed reference about the Pi camera options.

        :param listen_port: Port where the camera process will provide the video output while recording
        :type listen_port: int
        """

        super().__init__(**kwargs)

        self.listen_port = listen_port
        self.server_socket = socket.socket()
        self.server_socket.bind(('0.0.0.0', self.listen_port))
        self.server_socket.listen(0)

        import picamera

        self.camera = picamera.PiCamera()
        self.camera.resolution = (x_resolution, y_resolution)
        self.camera.framerate = framerate
        self.camera.hflip = hflip
        self.camera.vflip = vflip
        self.camera.sharpness = sharpness
        self.camera.contrast = contrast
        self.camera.brightness = brightness
        self.camera.video_stabilization = video_stabilization
        self.camera.ISO = iso
        self.camera.exposure_compensation = exposure_compensation
        self.camera.exposure_mode = exposure_mode
        self.camera.meter_mode = meter_mode
        self.camera.awb_mode = awb_mode
        self.camera.image_effect = image_effect
        self.camera.color_effects = color_effects
        self.camera.rotation = rotation
        self.camera.crop = crop
        self.start_recording_on_startup = start_recording_on_startup
        self.redis = None
        self.redis_queue = redis_queue
        self._recording_thread = None

    def send_camera_action(self, action, **kwargs):
        action = {
            'action': action.value,
            **kwargs
        }

        self.redis.send_message(msg=json.dumps(action), queue_name=self.redis_queue)

    def take_picture(self, image_file):
        """
        Take a picture.

        :param image_file: Output image file
        :type image_file: str
        """
        self.logger.info('Capturing camera snapshot to {}'.format(image_file))
        self.camera.capture(image_file)
        self.logger.info('Captured camera snapshot to {}'.format(image_file))

    # noinspection PyShadowingBuiltins
    def start_recording(self, video_file=None, format='h264'):
        """
        Start a recording.

        :param video_file: Output video file. If specified, the video will be recorded to file, otherwise it will be
            served via TCP/IP on the listen_port. Use ``stop_recording`` to stop the recording.
        :type video_file: str

        :param format: Video format (default: h264)
        :type format: str
        """

        # noinspection PyBroadException
        def recording_thread():
            if video_file:
                self.camera.start_recording(video_file, format=format)
                while True:
                    self.camera.wait_recording(2)
            else:
                while True:
                    connection = self.server_socket.accept()[0].makefile('wb')
                    self.logger.info('Accepted client connection on port {}'.format(self.listen_port))

                    try:
                        self.camera.start_recording(connection, format=format)
                        while True:
                            self.camera.wait_recording(2)
                    except ConnectionError:
                        self.logger.info('Client closed connection')
                        try:
                            self.stop_recording()
                        except:
                            pass

                        try:
                            connection.close()
                        except:
                            pass

                        self.send_camera_action(self.CameraAction.START_RECORDING)

        if self._recording_thread:
            self.logger.info('Recording already running')
            return

        self.logger.info('Starting camera recording')
        self._recording_thread = Thread(target=recording_thread,
                                        name='PiCameraRecorder')
        self._recording_thread.start()

    def stop_recording(self):
        """ Stops recording """

        self.logger.info('Stopping camera recording')

        try:
            self.camera.stop_recording()
        except Exception as e:
            self.logger.warning('Failed to stop recording')
            self.logger.exception(e)

    def run(self):
        super().run()

        if not self.redis:
            self.redis = get_backend('redis')

        if self.start_recording_on_startup:
            self.send_camera_action(self.CameraAction.START_RECORDING)

        self.logger.info('Initialized Pi camera backend')

        while not self.should_stop():
            try:
                msg = self.redis.get_message(self.redis_queue)

                if msg.get('action') == self.CameraAction.START_RECORDING:
                    self.start_recording()
                elif msg.get('action') == self.CameraAction.STOP_RECORDING:
                    self.stop_recording()
                elif msg.get('action') == self.CameraAction.TAKE_PICTURE:
                    self.take_picture(image_file=msg.get('image_file'))
            except Exception as e:
                self.logger.exception(e)


# vim:sw=4:ts=4:et:



import asyncio
import os
import ssl
import threading
import websockets

from platypush.backend import Backend
from platypush.context import get_plugin, get_or_create_event_loop
from platypush.message import Message
from platypush.message.request import Request
from platypush.message.response import Response
from platypush.utils import get_ssl_server_context


class WebsocketBackend(Backend):
    """
    Backend to communicate messages over a websocket medium.

    Requires:

        * **websockets** (``pip install websockets``)
    """

    # Websocket client message recv timeout in seconds
    _websocket_client_timeout = 0

    # Default websocket service port
    _default_websocket_port = 8765

    def __init__(self, port=_default_websocket_port, bind_address='0.0.0.0',
                 ssl_cafile=None, ssl_capath=None, ssl_cert=None, ssl_key=None,
                 client_timeout=_websocket_client_timeout, **kwargs):
        """
        :param port: Listen port for the websocket server (default: 8765)
        :type port: int

        :param bind_address: Bind address for the websocket server (default: 0.0.0.0, listen for any IP connection)
        :type websocket_port: str

        :param ssl_cert: Path to the certificate file if you want to enable SSL (default: None)
        :type ssl_cert: str

        :param ssl_key: Path to the key file if you want to enable SSL (default: None)
        :type ssl_key: str

        :param ssl_cafile: Path to the certificate authority file if required by the SSL configuration (default: None)
        :type ssl_cafile: str

        :param ssl_capath: Path to the certificate authority directory if required by the SSL configuration (default: None)
        :type ssl_capath: str

        :param client_timeout: Timeout without any messages being received before closing a client connection. A zero timeout keeps the websocket open until an error occurs (default: 0, no timeout)
        :type ping_timeout: int
        """

        super().__init__(**kwargs)

        self.port = port
        self.bind_address = bind_address
        self.client_timeout = client_timeout
        self.active_websockets = set()

        self.ssl_context = get_ssl_server_context(ssl_cert=ssl_cert,
                                                  ssl_key=ssl_key,
                                                  ssl_cafile=ssl_cafile,
                                                  ssl_capath=ssl_capath) \
            if ssl_cert else None

    def send_message(self, msg):
        websocket = get_plugin('websocket')
        websocket_args = {}

        if self.ssl_context:
            url = 'wss://localhost:{}'.format(self.port)
            websocket_args['ssl'] = self.ssl_context
        else:
            url = 'ws://localhost:{}'.format(self.port)

        websocket.send(url=url, msg=msg, **websocket_args)


    def notify_web_clients(self, event):
        """ Notify all the connected web clients (over websocket) of a new event """
        async def send_event(websocket):
            try:
                await websocket.send(str(event))
            except Exception as e:
                self.logger.warning('Error on websocket send_event: {}'.format(e))

        loop = get_or_create_event_loop()
        active_websockets = self.active_websockets.copy()

        for websocket in active_websockets:
            try:
                loop.run_until_complete(send_event(websocket))
            except websockets.exceptions.ConnectionClosed:
                self.logger.info('Client connection lost')
                self.active_websockets.remove(websocket)


    def run(self):
        super().run()

        async def serve_client(websocket, path):
            self.active_websockets.add(websocket)
            self.logger.debug('New websocket connection from {}'.
                             format(websocket.remote_address[0]))

            try:
                while True:
                    if self.client_timeout:
                        msg = await asyncio.wait_for(websocket.recv(),
                                                     timeout=self.client_timeout)
                    else:
                        msg = await websocket.recv()

                    msg = Message.build(msg)
                    self.logger.info('Received message from {}: {}'.
                                    format(websocket.remote_address[0], msg))

                    self.on_message(msg)

                    if isinstance(msg, Request):
                        response = self.get_message_response(msg) or Response()
                        self.logger.info('Processing response on the websocket backend: {}'.
                                         format(response))

                        await websocket.send(str(response))

            except websockets.exceptions.ConnectionClosed as e:
                self.active_websockets.remove(websocket)
                self.logger.debug('Websocket client {} closed connection'.
                                  format(websocket.remote_address[0]))
            except asyncio.TimeoutError as e:
                self.active_websockets.remove(websocket)
                self.logger.debug('Websocket connection to {} timed out'.
                                  format(websocket.remote_address[0]))
            except Exception as e:
                self.logger.exception(e)

        self.logger.info('Initialized websocket backend on port {}, bind address: {}'.
                         format(self.port, self.bind_address))

        websocket_args = {}
        if self.ssl_context:
            websocket_args['ssl'] = self.ssl_context

        loop = get_or_create_event_loop()
        server = websockets.serve(serve_client, self.bind_address, self.port,
                                  **websocket_args)

        loop.run_until_complete(server)
        loop.run_forever()


# vim:sw=4:ts=4:et:




import json
import logging
import time

from platypush.context import get_backend
from platypush.plugins import Plugin, action


class KafkaPlugin(Plugin):
    """
    Plugin to send messages to an Apache Kafka instance (https://kafka.apache.org/)

    Triggers:

        * :class:`platypush.message.event.kafka.KafkaMessageEvent` when a new message is received on the consumer topic.

    Requires:

        * **kafka** (``pip install kafka-python``)
    """

    def __init__(self, server=None, **kwargs):
        """
        :param server: Default Kafka server name or address + port (format: ``host:port``) to dispatch the messages to. If None (default), then it has to be specified upon message sending.
        :type server: str
        """

        super().__init__(**kwargs)

        self.server = '{server}:{port}'.format(server=server, port=port) \
            if server else None

        self.producer = None

        # Kafka can be veryyyy noisy
        logging.getLogger('kafka').setLevel(logging.ERROR)


    @action
    def send_message(self, msg, topic, server=None, **kwargs):
        """
        :param msg: Message to send - as a string, bytes stream, JSON, Platypush message, dictionary, or anything that implements ``__str__``

        :param server: Kafka server name or address + port (format: ``host:port``). If None, then the default server will be used
        :type server: str
        """

        from kafka import KafkaProducer

        if not server:
            if not self.server:
                try:
                    kafka_backend = get_backend('kafka')
                    server = kafka_backend.server
                except:
                    raise RuntimeError('No Kafka server nor default server specified')
            else:
                server = self.server

        if isinstance(msg, dict) or isinstance(msg, list):
            msg = json.dumps(msg)
        msg = str(msg).encode('utf-8')

        producer = KafkaProducer(bootstrap_servers=server)
        producer.send(topic, msg)
        producer.flush()


# vim:sw=4:ts=4:et:




from typing import Dict, List, Union, Optional

import numpy as np
from tensorflow.keras.models import Model

from platypush.message.response import Response


class TensorflowResponse(Response):
    """
    Generic Tensorflow response.
    """
    def __init__(self, *args, model: Model, **kwargs):
        """
        :param model: Name of the model.
        """
        super().__init__(*args, output={
            'model': model.name,
        }, **kwargs)

        self.model = model


class TensorflowTrainResponse(TensorflowResponse):
    """
    Tensorflow model fit/train response.
    """
    def __init__(self, *args, epochs: List[int], history: Dict[str, List[Union[int, float]]], **kwargs):
        """
        :param epochs: List of epoch indexes the model has been trained on.
        :param history: Train history, as a ``metric -> [values]`` dictionary where each value in ``values`` is
            the value for of that metric on a specific epoch.
        """
        super().__init__(*args, **kwargs)
        self.output['epochs'] = epochs
        self.output['history'] = history


class TensorflowPredictResponse(TensorflowResponse):
    """
    Tensorflow model prediction response.
    """
    def __init__(self, *args, prediction: np.ndarray, output_labels: Optional[List[str]] = None, **kwargs):
        super().__init__(*args, **kwargs)

        if output_labels and len(output_labels) == self.model.outputs[-1].shape[-1]:
            self.output['values'] = [
                {output_labels[i]: value for i, value in enumerate(p)}
                for p in prediction
            ]
        else:
            self.output['values'] = prediction

        if self.model.__class__.__name__ != 'LinearModel':
            prediction = [int(np.argmax(p)) for p in prediction]
            if output_labels:
                self.output['labels'] = [output_labels[p] for p in prediction]
            else:
                self.output['labels'] = prediction


# vim:sw=4:ts=4:et:



import logging
import threading
import time

from queue import Queue

from platypush.config import Config
from platypush.message.event import StopEvent

logger = logging.getLogger(__name__)


class Bus(object):
    """ Main local bus where the daemon will listen for new messages """

    _MSG_EXPIRY_TIMEOUT = 60.0  # Consider a message on the bus as expired after one minute without being picked up

    def __init__(self, on_message=None):
        self.bus = Queue()
        self.on_message = on_message
        self.thread_id = threading.get_ident()

    def post(self, msg):
        """ Sends a message to the bus """
        self.bus.put(msg)

    def get(self):
        """ Reads one message from the bus """
        return self.bus.get()

    def stop(self):
        """ Stops the bus by sending a STOP event """
        evt = StopEvent(target=Config.get('device_id'),
                        origin=Config.get('device_id'),
                        thread_id=self.thread_id)

        self.post(evt)

    def _msg_executor(self, msg):
        def executor():
            try:
                self.on_message(msg)
            except Exception as e:
                logger.error('Error on processing message {}'.format(msg))
                logger.exception(e)

        return executor

    def poll(self):
        """
        Reads messages from the bus until either stop event message or KeyboardInterrupt
        """

        if not self.on_message:
            logger.warning('No message handlers installed, cannot poll')
            return

        stop = False
        while not stop:
            msg = self.get()
            if msg.timestamp and time.time() - msg.timestamp > self._MSG_EXPIRY_TIMEOUT:
                logger.debug('{} seconds old message on the bus expired, ignoring it: {}'.
                             format(int(time.time()-msg.timestamp), msg))
                continue

            threading.Thread(target=self._msg_executor(msg)).start()

            if isinstance(msg, StopEvent) and msg.targets_me():
                logger.info('Received STOP event on the bus')
                stop = True


# vim:sw=4:ts=4:et:



import importlib

from enum import Enum
from threading import Timer
from time import time

from platypush.backend import Backend
from platypush.message.event.button.flic import FlicButtonEvent

from .fliclib.fliclib import FlicClient, ButtonConnectionChannel, ClickType


class ButtonFlicBackend(Backend):
    """
    Backend that listen for events from the Flic (https://flic.io/) bluetooth
    smart buttons.

    Triggers:

        * :class:`platypush.message.event.button.flic.FlicButtonEvent` when a button is pressed. The event will also contain the press sequence (e.g. ``["ShortPressEvent", "LongPressEvent", "ShortPressEvent"]``)

    Requires:

        * **fliclib** (https://github.com/50ButtonsEach/fliclib-linux-hci). For the backend to work properly you need to have the ``flicd`` daemon from the fliclib running, and you have to first pair the buttons with your device using any of the scanners provided by the library.
    """

    _long_press_timeout = 0.3
    _btn_timeout = 0.5
    ShortPressEvent = "ShortPressEvent"
    LongPressEvent = "LongPressEvent"

    def __init__(self, server='localhost', long_press_timeout=_long_press_timeout,
                 btn_timeout=_btn_timeout, **kwargs):
        """
        :param server: flicd server host (default: localhost)
        :type server: str

        :param long_press_timeout: How long you should press a button for a press action to be considered "long press" (default: 0.3 secohds)
        :type long_press_timeout: float

        :param btn_timeout: How long since the last button release before considering the user interaction completed (default: 0.5 seconds)
        :type btn_timeout: float
        """

        super().__init__(**kwargs)

        self.server = server
        self.client = FlicClient(server)
        self.client.get_info(self._received_info())
        self.client.on_new_verified_button = self._got_button()

        self._long_press_timeout = long_press_timeout
        self._btn_timeout = btn_timeout
        self._btn_timer = None
        self._btn_addr = None
        self._down_pressed_time = None
        self._cur_sequence = []

        self.logger.info('Initialized Flic buttons backend on {}'.format(self.server))

    def _got_button(self):
        def _f(bd_addr):
            cc = ButtonConnectionChannel(bd_addr)
            cc.on_button_up_or_down = \
                lambda channel, click_type, was_queued, time_diff: \
                self._on_event()(bd_addr, channel, click_type, was_queued, time_diff)
            self.client.add_connection_channel(cc)

        return _f

    def _received_info(self):
        def _f(items):
            for bd_addr in items["bd_addr_of_verified_buttons"]:
                self._got_button()(bd_addr)
        return _f

    def _on_btn_timeout(self):
        def _f():
            self.logger.info('Flic event triggered from {}: {}'.format(
                self._btn_addr, self._cur_sequence))

            self.bus.post(FlicButtonEvent(
                btn_addr=self._btn_addr, sequence=self._cur_sequence))

            self._cur_sequence = []

        return _f

    def _on_event(self):
        def _f(bd_addr, channel, click_type, was_queued, time_diff):
            if was_queued:
                return

            if self._btn_timer:
                self._btn_timer.cancel()

            if click_type == ClickType.ButtonDown:
                self._down_pressed_time = time()
                return

            btn_event = self.ShortPressEvent
            if self._down_pressed_time:
                if time() - self._down_pressed_time >= self._long_press_timeout:
                    btn_event = self.LongPressEvent
                self._down_pressed_time = None

            self._cur_sequence.append(btn_event)

            self._btn_addr = bd_addr
            self._btn_timer = Timer(self._btn_timeout, self._on_btn_timeout())
            self._btn_timer.start()

        return _f

    def run(self):
        super().run()

        self.client.handle_events()


# vim:sw=4:ts=4:et:




"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
.. license: MIT
"""

import json
import os
import time

from platypush.backend.assistant import AssistantBackend
from platypush.message.event.assistant import \
    ConversationStartEvent, ConversationEndEvent, ConversationTimeoutEvent, \
    ResponseEvent, NoResponseEvent, SpeechRecognizedEvent, AlarmStartedEvent, \
    AlarmEndEvent, TimerStartedEvent, TimerEndEvent, AlertStartedEvent, \
    AlertEndEvent, MicMutedEvent, MicUnmutedEvent


class AssistantGoogleBackend(AssistantBackend):
    """
    Google Assistant backend.

    It listens for voice commands and post conversation events on the bus.

    **WARNING**: This backend is deprecated, as the underlying Google Assistant
    library has been deprecated too: https://developers.google.com/assistant/sdk/reference/library/python/
    The old library might still work on some systems but its proper functioning
    is not guaranteed.
    Please use the Snowboy backend for hotword detection and the Google Assistant
    push-to-talk plugin for assistant interaction instead.

    Triggers:

        * :class:`platypush.message.event.assistant.ConversationStartEvent` \
            when a new conversation starts
        * :class:`platypush.message.event.assistant.SpeechRecognizedEvent` \
            when a new voice command is recognized
        * :class:`platypush.message.event.assistant.NoResponse` \
            when a conversation returned no response
        * :class:`platypush.message.event.assistant.ResponseEvent` \
            when the assistant is speaking a response
        * :class:`platypush.message.event.assistant.ConversationTimeoutEvent` \
            when a conversation times out
        * :class:`platypush.message.event.assistant.ConversationEndEvent` \
            when a new conversation ends
        * :class:`platypush.message.event.assistant.AlarmStartedEvent` \
            when an alarm starts
        * :class:`platypush.message.event.assistant.AlarmEndEvent` \
            when an alarm ends
        * :class:`platypush.message.event.assistant.TimerStartedEvent` \
            when a timer starts
        * :class:`platypush.message.event.assistant.TimerEndEvent` \
            when a timer ends
        * :class:`platypush.message.event.assistant.MicMutedEvent` \
            when the microphone is muted.
        * :class:`platypush.message.event.assistant.MicUnmutedEvent` \
            when the microphone is un-muted.

    Requires:

        * **google-assistant-library** (``pip install google-assistant-library``)
        * **google-assistant-sdk[samples]** (``pip install google-assistant-sdk[samples]``)
    """

    def __init__(self,
                 credentials_file=os.path.join(
                     os.path.expanduser('~/.config'),
                     'google-oauthlib-tool', 'credentials.json'),
                 device_model_id='Platypush', **kwargs):
        """
        :param credentials_file: Path to the Google OAuth credentials file \
            (default: ~/.config/google-oauthlib-tool/credentials.json). \
            See https://developers.google.com/assistant/sdk/guides/library/python/embed/install-sample#generate_credentials \
            for instructions to get your own credentials file.

        :type credentials_file: str

        :param device_model_id: Device model ID to use for the assistant \
            (default: Platypush)
        :type device_model_id: str
        """

        super().__init__(**kwargs)
        self.credentials_file = credentials_file
        self.device_model_id = device_model_id
        self.credentials = None
        self.assistant = None
        self._has_error = False
        self._is_muted = False

        self.logger.info('Initialized Google Assistant backend')

    def _process_event(self, event):
        from google.assistant.library.event import EventType, AlertType

        self.logger.info('Received assistant event: {}'.format(event))
        self._has_error = False

        if event.type == EventType.ON_CONVERSATION_TURN_STARTED:
            self.bus.post(ConversationStartEvent(assistant=self))
        elif event.type == EventType.ON_CONVERSATION_TURN_FINISHED:
            if not event.args.get('with_follow_on_turn'):
                self.bus.post(ConversationEndEvent(assistant=self))
        elif event.type == EventType.ON_CONVERSATION_TURN_TIMEOUT:
            self.bus.post(ConversationTimeoutEvent(assistant=self))
        elif event.type == EventType.ON_NO_RESPONSE:
            self.bus.post(NoResponseEvent(assistant=self))
        elif hasattr(EventType, 'ON_RENDER_RESPONSE') and \
                event.type == EventType.ON_RENDER_RESPONSE:
            self.bus.post(ResponseEvent(assistant=self, response_text=event.args.get('text')))
            tts, args = self._get_tts_plugin()

            if tts and 'text' in event.args:
                self.stop_conversation()
                tts.say(text=event.args['text'], **args)
        elif hasattr(EventType, 'ON_RESPONDING_STARTED') and \
                event.type == EventType.ON_RESPONDING_STARTED and \
                event.args.get('is_error_response', False) is True:
            self.logger.warning('Assistant response error')
        elif event.type == EventType.ON_RECOGNIZING_SPEECH_FINISHED:
            phrase = event.args['text'].lower().strip()
            self.logger.info('Speech recognized: {}'.format(phrase))
            self.bus.post(SpeechRecognizedEvent(assistant=self, phrase=phrase))
        elif event.type == EventType.ON_ALERT_STARTED:
            if event.args.get('alert_type') == AlertType.ALARM:
                self.bus.post(AlarmStartedEvent(assistant=self))
            elif event.args.get('alert_type') == AlertType.TIMER:
                self.bus.post(TimerStartedEvent(assistant=self))
            else:
                self.bus.post(AlertStartedEvent(assistant=self))
        elif event.type == EventType.ON_ALERT_FINISHED:
            if event.args.get('alert_type') == AlertType.ALARM:
                self.bus.post(AlarmEndEvent(assistant=self))
            elif event.args.get('alert_type') == AlertType.TIMER:
                self.bus.post(TimerEndEvent(assistant=self))
            else:
                self.bus.post(AlertEndEvent(assistant=self))
        elif event.type == EventType.ON_ASSISTANT_ERROR:
            self._has_error = True
            if event.args.get('is_fatal'):
                self.logger.error('Fatal assistant error')
            else:
                self.logger.warning('Assistant error')
        if event.type == EventType.ON_MUTED_CHANGED:
            self._is_muted = event.args.get('is_muted')
            event = MicMutedEvent() if self._is_muted else MicUnmutedEvent()
            self.bus.post(event)

    def start_conversation(self):
        """ Starts an assistant conversation """
        if self.assistant:
            self.assistant.start_conversation()

    def stop_conversation(self):
        """ Stops an assistant conversation """
        if self.assistant:
            self.assistant.stop_conversation()

    def set_mic_mute(self, muted):
        if not self.assistant:
            self.logger.warning('Assistant not running')
            return

        self.assistant.set_mic_mute(muted)

    def is_muted(self) -> bool:
        return self._is_muted

    def send_text_query(self, query):
        if not self.assistant:
            self.logger.warning('Assistant not running')
            return

        self.assistant.send_text_query(query)

    def run(self):
        import google.oauth2.credentials
        from google.assistant.library import Assistant

        super().run()

        with open(self.credentials_file, 'r') as f:
            self.credentials = google.oauth2.credentials.Credentials(token=None, **json.load(f))

        while not self.should_stop():
            self._has_error = False

            with Assistant(self.credentials, self.device_model_id) as assistant:
                self.assistant = assistant
                for event in assistant.start():
                    if not self.is_detecting():
                        self.logger.info('Assistant event received but detection is currently paused')
                        continue

                    self._process_event(event)
                    if self._has_error:
                        self.logger.info('Restarting the assistant after an unrecoverable error')
                        time.sleep(5)
                        break


# vim:sw=4:ts=4:et:



from typing import Optional, Dict, Any

from platypush.plugins import Plugin, action
from platypush.message.response.weather.buienradar import BuienradarWeatherResponse, BuienradarPrecipitationResponse, \
    BuienradarForecastResponse, BuienradarForecast


class WeatherBuienradarPlugin(Plugin):
    """
    Plugin for getting weather updates through Buienradar - a Dutch weather app.

    Requires:

        * **buienradar** (``pip install buienradar``)

    """

    def __init__(self, lat: float, long: float, time_frame: int = 120, **kwargs):
        """
        :param lat: Default latitude
        :param long: Default longitude
        :param time_frame: Default number of minutes to look ahead for precipitation forecast
        """
        super().__init__(**kwargs)
        self.lat = lat
        self.long = long
        self.time_frame = time_frame
        self.latest_bulletin = {}

    def get_data(self, lat: Optional[float] = None, long: Optional[float] = None, time_frame: Optional[int] = None) \
            -> Dict[str, Any]:
        # noinspection PyPackageRequirements
        from buienradar.buienradar import get_data, parse_data
        # noinspection PyPackageRequirements
        from buienradar.constants import SUCCESS, CONTENT, RAINCONTENT, DATA

        lat = lat or self.lat
        long = long or self.long
        time_frame = time_frame or self.time_frame

        result = get_data(latitude=lat, longitude=long)
        if not result.get(SUCCESS):
            raise RuntimeError('Error while retrieving data')

        data = result.get(CONTENT)
        rain_data = result.get(RAINCONTENT)
        result = parse_data(data, rain_data, lat, long, time_frame)
        return result.get(DATA, {})

    @action
    def get_weather(self, lat: Optional[float] = None, long: Optional[float] = None) -> BuienradarWeatherResponse:
        """
        Get the current weather conditions.

        :param lat: Weather latitude (default: configured latitude)
        :param long: Weather longitude (default: configured longitude)
        """
        data = self.get_data(lat, long, 60)

        return BuienradarWeatherResponse(
            barometer_fc=data.get('barometerfcname'),
            condition_name=data.get('condition', {}).get('condition'),
            condition_name_long=data.get('condition', {}).get('exact'),
            condition_image=data.get('condition', {}).get('image'),
            feel_temperature=data.get('feeltemperature'),
            ground_temperature=data.get('groundtemperature'),
            humidity=data.get('humidity'),
            irradiance=data.get('irradiance'),
            measured=data.get('measured'),
            precipitation=data.get('precipitation'),
            pressure=data.get('pressure'),
            rain_last_24_hours=data.get('rainlast24hour'),
            rain_last_hour=data.get('rainlasthour'),
            station_name=data.get('stationname'),
            temperature=data.get('temperature'),
            visibility=data.get('visibility'),
            wind_azimuth=data.get('windazimuth'),
            wind_direction=data.get('wind_irection'),
            wind_force=data.get('windforce'),
            wind_gust=data.get('windgust'),
            wind_speed=data.get('windspeed')
        )

    @action
    def get_forecast(self, lat: Optional[float] = None, long: Optional[float] = None) -> BuienradarForecastResponse:
        """
        Get the weather forecast for the next days.

        :param lat: Weather latitude (default: configured latitude)
        :param long: Weather longitude (default: configured longitude)
        """
        data = self.get_data(lat, long, 60).get('forecast', [])
        return BuienradarForecastResponse([
            BuienradarForecast(
                condition_name=d.get('condition', {}).get('condition'),
                condition_name_long=d.get('condition', {}).get('exact'),
                condition_image=d.get('condition', {}).get('image'),
                date_time=d.get('datetime'),
                rain=d.get('rain'),
                min_rain=d.get('minrain'),
                max_rain=d.get('maxrain'),
                rain_chance=d.get('rainchance'),
                snow=d.get('snow'),
                temperature=d.get('temperature'),
                wind_azimuth=d.get('windazimuth'),
                wind_direction=d.get('winddirection'),
                wind_force=d.get('windforce'),
                wind_speed=d.get('windspeed'),
            )
            for d in data
        ])

    @action
    def get_precipitation(self, lat: Optional[float] = None, long: Optional[float] = None,
                          time_frame: Optional[int] = None) -> BuienradarPrecipitationResponse:
        """
        Get the precipitation forecast for the specified time frame.

        :param lat: Weather latitude (default: configured latitude)
        :param long: Weather longitude (default: configured longitude)
        :param time_frame: Time frame for the forecast in minutes (default: configured time_frame)
        """
        data = self.get_data(lat, long, time_frame).get('precipitation_forecast', {})
        return BuienradarPrecipitationResponse(
            average=data.get('average'),
            total=data.get('total'),
            time_frame=data.get('timeframe'),
        )


# vim:sw=4:ts=4:et:



from platypush.plugins import Plugin, action


class SensorPlugin(Plugin):
    """
    Sensor abstract plugin. Any plugin that interacts with sensors
    should implement this class (and the get_measurement() method)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @action
    def get_measurement(self, *args, **kwargs):
        """
        Implemented by the subclasses.

        :returns: Either a raw scalar:

            ``output = 273.16``

        or a name-value dictionary with the values that have been read::

            output = {
                "temperature": 21.5,
                "humidity": 41.0
            }

        or a list of values::

            [
                0.01,
                0.34,
                0.53,
                ...
            ]

        """
        raise NotImplementedError('get_measurement should be implemented in a derived class')

    @action
    def get_data(self, *args, **kwargs):
        """
        Alias for ``get_measurement``
        """

        return self.get_measurement(*args, **kwargs).output

    @action
    def close(self):
        pass


# vim:sw=4:ts=4:et:



import json
import os
import random
import threading
import time

import urllib.request
import urllib.parse

from platypush.context import get_bus
from platypush.plugins import Plugin, action
from platypush.message.event.torrent import \
    TorrentDownloadStartEvent, TorrentDownloadedMetadataEvent, TorrentStateChangeEvent, \
    TorrentDownloadProgressEvent, TorrentDownloadCompletedEvent, TorrentDownloadStopEvent, \
    TorrentPausedEvent, TorrentResumedEvent, TorrentQueuedEvent


class TorrentPlugin(Plugin):
    """
    Plugin to search and download torrents.

    Requires:

        * **python-libtorrent** (``pip install git+https://github.com/arvidn/libtorrent``)
        * **requests** (``pip install requests``) [optional] for torrent info URL download
    """

    # Wait time in seconds between two torrent transfer checks
    _MONITOR_CHECK_INTERVAL = 3

    default_torrent_ports = [6881, 6891]
    categories = {
        'movies': None,
        'tv': None,
        'anime': None,
    }

    torrent_state = {}
    transfers = {}
    headers = {
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.103 Safari/537.36'
    }

    def __init__(self, download_dir=None, torrent_ports=None, **kwargs):
        """
        :param download_dir: Directory where the videos/torrents will be downloaded (default: none)
        :type download_dir: str

        :param torrent_ports: Torrent ports to listen on (default: 6881 and 6891)
        :type torrent_ports: list[int]
        """

        super().__init__(**kwargs)

        if torrent_ports is None:
            torrent_ports = []

        for category in self.categories.keys():
            self.categories[category] = getattr(self, 'search_' + category)

        self.torrent_ports = torrent_ports if torrent_ports else self.default_torrent_ports
        self.download_dir = None
        self._session = None

        if download_dir:
            self.download_dir = os.path.abspath(os.path.expanduser(download_dir))

    # noinspection PyIncorrectDocstring
    @action
    def search(self, query, category=None, *args, **kwargs):
        """
        Perform a search of video torrents.

        :param query: Query string, video name or partial name
        :type query: str

        :param category: Category to search. Supported types: "movies", "tv", "anime".
            Default: None (search all categories)
        :type category: str or list

        :param language: Language code for the results - example: "en" (default: None, no filter)
        :type language: str
        """

        if isinstance(category, str):
            category = [category]

        results = {
            category: []
            for category in (category or self.categories.keys())
        }

        # noinspection PyCallingNonCallable
        def worker(cat):
            if cat not in self.categories:
                raise RuntimeError('Unsupported category {}. Supported category: {}'.
                                   format(cat, self.categories.keys()))

            self.logger.info('Searching {} torrents for "{}"'.format(cat, query))
            results[cat] = self.categories[cat](query, *args, **kwargs)

        workers = [
            threading.Thread(target=worker, kwargs={'cat': category})
            for category in (category or self.categories.keys())
        ]

        for worker in workers:
            worker.start()

        for worker in workers:
            worker.join()

        ret = []
        for (category, items) in results.items():
            ret += items
        return ret

    def search_movies(self, query, language=None):
        request = urllib.request.urlopen(urllib.request.Request(
            'https://popcorn-api.io/movies/1?' + urllib.parse.urlencode({
                'sort': 'relevance',
                'keywords': query,
            }), headers=self.headers
        ))

        response = json.loads(request.read().decode('utf-8'))

        return sorted([
            {
                'imdb_id': result.get('imdb_id'),
                'type': 'movies',
                'title': '{title} [movies][{language}][{quality}]'.format(
                    title=result.get('title'), language=lang, quality=quality),
                'year': result.get('year'),
                'synopsis': result.get('synopsis'),
                'trailer': result.get('trailer'),
                'language': lang,
                'quality': quality,
                'size': item.get('size'),
                'seeds': item.get('seed'),
                'peers': item.get('peer'),
                'url': item.get('url'),
            }
            for result in (response or [])
            for (lang, items) in result.get('torrents', {}).items()
            if not language or language == lang
            for (quality, item) in items.items()
        ], key=lambda _: _.get('seeds'), reverse=True)

    def search_tv(self, query):
        request = urllib.request.urlopen(urllib.request.Request(
            'https://popcorn-api.io/shows/1?' + urllib.parse.urlencode({
                'sort': 'relevance',
                'keywords': query,
            }), headers=self.headers
        ))

        results = []
        shows = json.loads(request.read().decode('utf-8'))

        for show in shows:
            request = urllib.request.urlopen(urllib.request.Request(
                'https://tv-v2.api-fetch.website/show/' + show.get('_id'),
                headers=self.headers))

            show = json.loads(request.read().decode('utf-8'))

            results.extend(sorted([
                {
                    'imdb_id': show.get('imdb_id'),
                    'tvdb_id': show.get('tvdb_id'),
                    'type': 'tv',
                    'title': '[{show}][s{season}e{episode:02d}] {title} [tv][{quality}]'.format(
                        show=show.get('title'), season=episode.get('season', 0),
                        episode=episode.get('episode', 0), title=episode.get('title'), quality=quality),
                    'season': episode.get('season'),
                    'episode': episode.get('episode'),
                    'year': show.get('year'),
                    'first_aired': episode.get('first_aired'),
                    'quality': quality,
                    'num_seasons': show.get('num_seasons'),
                    'status': show.get('status'),
                    'network': show.get('network'),
                    'country': show.get('country'),
                    'show_synopsis': show.get('synopsys'),
                    'synopsis': episode.get('overview'),
                    'provider': torrent.get('provider'),
                    'seeds': torrent.get('seeds'),
                    'peers': torrent.get('peers'),
                    'url': torrent.get('url'),
                }
                for episode in show.get('episodes', [])
                for quality, torrent in episode.get('torrents', {}).items()
                if quality != '0'
            ], key=lambda _: int(_.get('season')*100) + int(_.get('episode'))))

        return results

    def search_anime(self, query):
        request = urllib.request.urlopen(urllib.request.Request(
            'https://popcorn-api.io/animes/1?' + urllib.parse.urlencode({
                'sort': 'relevance',
                'keywords': query,
            }), headers=self.headers
        ))

        results = []
        shows = json.loads(request.read().decode('utf-8'))

        for show in shows:
            request = urllib.request.urlopen(urllib.request.Request(
                'https://anime.api-fetch.website/anime/' + show.get('_id'),
                headers=self.headers))

            show = json.loads(request.read().decode('utf-8'))

            results.extend(sorted([
                {
                    'tvdb_id': episode.get('tvdb_id'),
                    'type': 'anime',
                    'title': '[{show}][s{season}e{episode:02d}] {title} [anime][{quality}]'.format(
                        show=show.get('title'), season=int(episode.get('season', 0)),
                        episode=int(episode.get('episode', 0)), title=episode.get('title'), quality=quality),
                    'season': episode.get('season'),
                    'episode': episode.get('episode'),
                    'year': show.get('year'),
                    'quality': quality,
                    'num_seasons': show.get('num_seasons'),
                    'status': show.get('status'),
                    'show_synopsis': show.get('synopsys'),
                    'synopsis': episode.get('overview'),
                    'seeds': torrent.get('seeds'),
                    'peers': torrent.get('peers'),
                    'url': torrent.get('url'),
                }
                for episode in show.get('episodes', [])
                for quality, torrent in episode.get('torrents', {}).items()
                if quality != '0'
            ], key=lambda _: int(_.get('season')*100) + int(_.get('episode'))))

        return results

    # noinspection PyArgumentList
    def _get_torrent_info(self, torrent, download_dir):
        import libtorrent as lt

        torrent_file = None
        magnet = None
        info = {}
        file_info = {}

        if torrent.startswith('magnet:?'):
            magnet = torrent
            info = lt.parse_magnet_uri(magnet)
            info['magnet'] = magnet
        elif torrent.startswith('http://') or torrent.startswith('https://'):
            import requests
            response = requests.get(torrent, allow_redirects=True)
            torrent_file = os.path.join(download_dir,
                                        self._generate_rand_filename())

            with open(torrent_file, 'wb') as f:
                f.write(response.content)
        else:
            torrent_file = os.path.abspath(os.path.expanduser(torrent))
            if not os.path.isfile(torrent_file):
                raise RuntimeError('{} is not a valid torrent file'.format(torrent_file))

        if torrent_file:
            file_info = lt.torrent_info(torrent_file)
            info = {
                'name': file_info.name(),
                'url': torrent,
                'trackers': [t.url for t in list(file_info.trackers())],
                'save_path': download_dir,
            }

        return info, file_info, torrent_file, magnet

    def _fire_event(self, event, event_hndl):
        bus = get_bus()
        bus.post(event)

        try:
            if event_hndl:
                event_hndl(event)
        except Exception as e:
            self.logger.warning('Exception in torrent event handler: {}'.format(str(e)))
            self.logger.exception(e)

    def _torrent_monitor(self, torrent, transfer, download_dir, event_hndl, is_media):
        def thread():
            files = []
            last_status = None
            download_started = False
            metadata_downloaded = False

            while not transfer.is_finished():
                if torrent not in self.transfers:
                    self.logger.info('Torrent {} has been stopped and removed'.format(torrent))
                    self._fire_event(TorrentDownloadStopEvent(url=torrent), event_hndl)
                    break

                status = transfer.status()
                torrent_file = transfer.torrent_file()

                if torrent_file:
                    self.torrent_state[torrent]['size'] = torrent_file.total_size()
                    files = [os.path.join(
                        download_dir,
                        torrent_file.files().file_path(i))
                        for i in range(0, torrent_file.files().num_files())
                    ]

                    if is_media:
                        from platypush.plugins.media import MediaPlugin
                        # noinspection PyProtectedMember
                        files = [f for f in files if MediaPlugin._is_video_file(f)]

                self.torrent_state[torrent]['download_rate'] = status.download_rate
                self.torrent_state[torrent]['name'] = status.name
                self.torrent_state[torrent]['num_peers'] = status.num_peers
                self.torrent_state[torrent]['paused'] = status.paused
                self.torrent_state[torrent]['progress'] = round(100 * status.progress, 2)
                self.torrent_state[torrent]['state'] = status.state.name
                self.torrent_state[torrent]['title'] = status.name
                self.torrent_state[torrent]['torrent'] = torrent
                self.torrent_state[torrent]['upload_rate'] = status.upload_rate
                self.torrent_state[torrent]['url'] = torrent
                self.torrent_state[torrent]['files'] = files

                if transfer.has_metadata() and not metadata_downloaded:
                    self._fire_event(TorrentDownloadedMetadataEvent(**self.torrent_state[torrent]), event_hndl)
                    metadata_downloaded = True

                if status.state == status.downloading and not download_started:
                    self._fire_event(TorrentDownloadStartEvent(**self.torrent_state[torrent]), event_hndl)
                    download_started = True

                if last_status and status.progress != last_status.progress:
                    self._fire_event(TorrentDownloadProgressEvent(**self.torrent_state[torrent]), event_hndl)

                if not last_status or status.state != last_status.state:
                    self._fire_event(TorrentStateChangeEvent(**self.torrent_state[torrent]), event_hndl)

                if last_status and status.paused != last_status.paused:
                    if status.paused:
                        self._fire_event(TorrentPausedEvent(**self.torrent_state[torrent]), event_hndl)
                    else:
                        self._fire_event(TorrentResumedEvent(**self.torrent_state[torrent]), event_hndl)

                last_status = status
                time.sleep(self._MONITOR_CHECK_INTERVAL)

            if transfer and transfer.is_finished():
                self._fire_event(TorrentDownloadCompletedEvent(**self.torrent_state[torrent]), event_hndl)

            self.remove(torrent)
            return files

        return thread

    @action
    def download(self, torrent, download_dir=None, _async=False, event_hndl=None, is_media=False):
        """
        Download a torrent.

        :param torrent: Torrent to download. Supported formats:

            * Magnet URLs
            * Torrent URLs
            * Local torrent file

        :type torrent: str

        :param download_dir: Directory to download, overrides the default download_dir attribute (default: None)
        :type download_dir: str

        :param _async: If true then the method will add the torrent to the transfer and then return. Updates on
            the download status should be retrieved either by listening to torrent events or registering the
            event handler. If false (default) then the method will exit only when the torrent download is complete.
        :type _async: bool

        :param event_hndl: A function that takes an event object as argument and is invoked upon a new torrent event
            (download started, progressing, completed etc.)
        :type event_hndl: function

        :param is_media: Set it to true if you're downloading a media file that you'd like to stream as soon as the
            first chunks are available. If so, then the events and the status method will only include media files
        :type is_media: bool
        """

        if torrent in self.torrent_state and torrent in self.transfers:
            return self.torrent_state[torrent]

        import libtorrent as lt

        if not download_dir:
            if self.download_dir:
                download_dir = self.download_dir
            else:
                raise RuntimeError('download_dir not specified')

        download_dir = os.path.abspath(os.path.expanduser(download_dir))
        os.makedirs(download_dir, exist_ok=True)
        info, file_info, torrent_file, magnet = self._get_torrent_info(torrent, download_dir)

        if not self._session:
            # noinspection PyArgumentList
            self._session = lt.session()
            self._session.listen_on(*self.torrent_ports)

        params = {
            'save_path': download_dir,
            'storage_mode': lt.storage_mode_t.storage_mode_sparse,
        }

        if magnet:
            transfer = lt.add_magnet_uri(self._session, magnet, params)
        else:
            params['ti'] = file_info
            transfer = self._session.add_torrent(params)

        self.transfers[torrent] = transfer
        self.torrent_state[torrent] = {
            'url': torrent,
            'title': transfer.status().name,
            'trackers': info['trackers'],
            'save_path': download_dir,
        }

        self._fire_event(TorrentQueuedEvent(url=torrent), event_hndl)
        self.logger.info('Downloading "{}" to "{}" from [{}]'.format(info['name'], download_dir, torrent))
        monitor_thread = self._torrent_monitor(torrent=torrent, transfer=transfer, download_dir=download_dir,
                                               event_hndl=event_hndl, is_media=is_media)

        if not _async:
            return monitor_thread()

        threading.Thread(target=monitor_thread).start()
        return self.torrent_state[torrent]

    @action
    def status(self, torrent=None):
        """
        Get the status of the current transfers.

        :param torrent: Torrent path, URL or magnet URI whose status will be retrieve (default: None, retrieve all
            current transfers)
        :type torrent: str

        :returns: A dictionary in the format torrent_url -> status
        """

        if torrent:
            return self.torrent_state.get(torrent)
        return self.torrent_state

    @action
    def pause(self, torrent):
        """
        Pause/resume a torrent transfer.

        :param torrent: Torrent URL as returned from `get_status()`
        :type torrent: str
        """

        if torrent not in self.transfers:
            return None, "No transfer in progress for {}".format(torrent)

        if self.torrent_state[torrent].get('paused', False):
            self.transfers[torrent].resume()
        else:
            self.transfers[torrent].pause()

        return self.torrent_state[torrent]

    @action
    def resume(self, torrent):
        """
        Resume a torrent transfer.

        :param torrent: Torrent URL as returned from `get_status()`
        :type torrent: str
        """

        if torrent not in self.transfers:
            return None, "No transfer in progress for {}".format(torrent)

        self.transfers[torrent].resume()

    @action
    def remove(self, torrent):
        """
        Stops and removes a torrent transfer.

        :param torrent: Torrent URL as returned from `get_status()`
        :type torrent: str
        """

        if torrent not in self.transfers:
            return None, "No transfer in progress for {}".format(torrent)

        self.transfers[torrent].pause()
        del self.torrent_state[torrent]
        del self.transfers[torrent]

        if not len(self.transfers):
            self.logger.info('No remaining active torrent transfers found, exiting session')
            self._session = None

    @action
    def quit(self):
        """
        Quits all the transfers and the active session
        """
        if not self._session:
            self.logger.info('No active sessions found')
            return

        transfers = self.transfers.copy()
        for torrent in transfers:
            self.remove(torrent)

    @staticmethod
    def _generate_rand_filename(length=16):
        name = ''
        for i in range(0, length):
            name += hex(random.randint(0, 15))[2:].upper()
        return name + '.torrent'


# vim:sw=4:ts=4:et:



import os
import subprocess
import threading

from multiprocessing import Process

from platypush.backend import Backend
from platypush.backend.http.app import application
from platypush.context import get_or_create_event_loop
from platypush.utils import get_ssl_server_context, set_thread_name


class HttpBackend(Backend):
    """
    The HTTP backend is a general-purpose web server that you can leverage:

        * To execute Platypush commands via HTTP calls. Example::

            curl -XPOST -H 'Content-Type: application/json' -H "X-Token: your_token" \\
                -d '{
                    "type":"request",
                    "target":"nodename",
                    "action":"tts.say",
                    "args": {"phrase":"This is a test"}
                }' \\
                http://localhost:8008/execute

        * To interact with your system (and control plugins and backends) through the Platypush web panel,
          by default available on your web root document. Any plugin that you have configured and available as a panel
          plugin will appear on the web panel as well as a tab.

        * To display a fullscreen dashboard with your configured widgets, by default available under ``/dashboard``

        * To stream media over HTTP through the ``/media`` endpoint

    Any plugin can register custom routes under ``platypush/backend/http/app/routes/plugins``.
    Any additional route is managed as a Flask blueprint template and the `.py`
    module can expose lists of routes to the main webapp through the
    ``__routes__`` object (a list of Flask blueprints).

    Note that if you set up a main token, it will be required for any HTTP
    interaction - either as ``X-Token`` HTTP header, on the query string
    (attribute name: ``token``), as part of the JSON payload root (attribute
    name: ``token``), or via HTTP basic auth (any username works).

    Requires:

        * **flask** (``pip install flask``)
        * **redis** (``pip install redis``)
        * **websockets** (``pip install websockets``)
        * **python-dateutil** (``pip install python-dateutil``)
        * **magic** (``pip install python-magic``), optional, for MIME type
            support if you want to enable media streaming
        * **uwsgi** (``pip install uwsgi`` plus uwsgi server installed on your
            system if required) - optional but recommended. By default the
            Platypush web server will run in a process spawned on the fly by
            the HTTP backend. However, being a Flask app, it will serve clients
            in a single thread and won't support many features of a full-blown
            web server.

    Base command to run the web server over uwsgi::

        uwsgi --http :8008 --module platypush.backend.http.uwsgi --master --processes 4 --threads 4

    Bear in mind that the main webapp is defined in ``platypush.backend.http.app:application``
    and the WSGI startup script is stored under ``platypush/backend/http/uwsgi.py``.
    """

    _DEFAULT_HTTP_PORT = 8008
    _DEFAULT_WEBSOCKET_PORT = 8009

    def __init__(self, port=_DEFAULT_HTTP_PORT,
                 websocket_port=_DEFAULT_WEBSOCKET_PORT,
                 disable_websocket=False, dashboard=None, resource_dirs=None,
                 ssl_cert=None, ssl_key=None, ssl_cafile=None, ssl_capath=None,
                 maps=None, run_externally=False, uwsgi_args=None, **kwargs):
        """
        :param port: Listen port for the web server (default: 8008)
        :type port: int

        :param websocket_port: Listen port for the websocket server (default: 8009)
        :type websocket_port: int

        :param disable_websocket: Disable the websocket interface (default: False)
        :type disable_websocket: bool

        :param ssl_cert: Set it to the path of your certificate file if you want to enable HTTPS (default: None)
        :type ssl_cert: str

        :param ssl_key: Set it to the path of your key file if you want to enable HTTPS (default: None)
        :type ssl_key: str

        :param ssl_cafile: Set it to the path of your certificate authority file if you want to enable HTTPS
            (default: None)
        :type ssl_cafile: str

        :param ssl_capath: Set it to the path of your certificate authority directory if you want to enable HTTPS
            (default: None)
        :type ssl_capath: str

        :param resource_dirs: Static resources directories that will be
            accessible through ``/resources/<path>``. It is expressed as a map
            where the key is the relative path under ``/resources`` to expose and
            the value is the absolute path to expose.
        :type resource_dirs: dict[str, str]

        :param dashboard: Set it if you want to use the dashboard service. It will contain the configuration for the
            widgets to be used (look under ``platypush/backend/http/templates/widgets/`` for the available widgets).

        Example configuration::

            dashboard:
                background_image: https://site/image.png
                widgets:                # Each row of the dashboard will have 6 columns
                    -
                        widget: calendar           # Calendar widget
                        columns: 6
                    -
                        widget: music              # Music widget
                        columns: 3
                    -
                        widget: date-time-weather  # Date, time and weather widget
                        columns: 3
                    -
                        widget: image-carousel     # Image carousel
                        columns: 6
                        # Absolute path (valid as long as it's a subdirectory of one of the available `resource_dirs`)
                        images_path: ~/Dropbox/Photos/carousel
                        refresh_seconds: 15
                    -
                        widget: rss-news           # RSS feeds widget
                        # Requires backend.http.poll to be enabled with some RSS sources and write them to sqlite db
                        columns: 6
                        limit: 25
                        db: "sqlite:////home/user/.local/share/platypush/feeds/rss.db"

        :type dashboard: dict

        :param run_externally: If set, then the HTTP backend will not directly
            spawn the web server. Set this option if you plan to run the webapp
            in a separate web server (recommended), like uwsgi or uwsgi+nginx.
        :type run_externally: bool

        :param uwsgi_args: If ``run_externally`` is set and you would like the
            HTTP backend to directly spawn and control the uWSGI application
            server instance, then pass the list of uWSGI arguments through
            this parameter. Some examples include::

                # Start uWSGI instance listening on HTTP port 8008 with 4
                # processes
                ['--plugin', 'python', '--http-socket', ':8008', '--master', '--processes', '4']

                # Start uWSGI instance listening on uWSGI socket on port 3031.
                # You can then use another full-blown web server, like nginx
                # or Apache, to communicate with the uWSGI instance
                ['--plugin', 'python', '--socket', '127.0.0.1:3031', '--master', '--processes', '4']
        :type uwsgi_args: list[str]
        """

        super().__init__(**kwargs)

        self.port = port
        self.websocket_port = websocket_port
        self.dashboard = dashboard or {}
        self.maps = maps or {}
        self.server_proc = None
        self.disable_websocket = disable_websocket
        self.websocket_thread = None

        if resource_dirs:
            self.resource_dirs = {name: os.path.abspath(
                os.path.expanduser(d)) for name, d in resource_dirs.items()}
        else:
            self.resource_dirs = {}

        self.active_websockets = set()
        self.run_externally = run_externally
        self.uwsgi_args = uwsgi_args or []
        self.ssl_context = get_ssl_server_context(ssl_cert=ssl_cert,
                                                  ssl_key=ssl_key,
                                                  ssl_cafile=ssl_cafile,
                                                  ssl_capath=ssl_capath) \
            if ssl_cert else None

        if self.uwsgi_args:
            self.uwsgi_args = [str(_) for _ in self.uwsgi_args] + \
                ['--module', 'platypush.backend.http.uwsgi', '--enable-threads']

        self.local_base_url = '{proto}://localhost:{port}'.\
            format(proto=('https' if ssl_cert else 'http'), port=self.port)

        self._websocket_lock_timeout = 10
        self._websocket_lock = threading.RLock()
        self._websocket_locks = {}

    def send_message(self, msg, **kwargs):
        self.logger.warning('Use cURL or any HTTP client to query the HTTP backend')

    def on_stop(self):
        """ On backend stop """
        self.logger.info('Received STOP event on HttpBackend')

        if self.server_proc:
            if isinstance(self.server_proc, subprocess.Popen):
                self.server_proc.kill()
                self.server_proc.wait()
            else:
                self.server_proc.terminate()
                self.server_proc.join()

    def _acquire_websocket_lock(self, ws):
        try:
            acquire_ok = self._websocket_lock.acquire(timeout=self._websocket_lock_timeout)
            if not acquire_ok:
                raise TimeoutError('Websocket lock acquire timeout')

            addr = ws.remote_address
            if addr not in self._websocket_locks:
                self._websocket_locks[addr] = threading.RLock()
        finally:
            self._websocket_lock.release()

        acquire_ok = self._websocket_locks[addr].acquire(timeout=self._websocket_lock_timeout)
        if not acquire_ok:
            raise TimeoutError('Websocket on address {} not ready to receive data'.format(addr))

    def _release_websocket_lock(self, ws):
        try:
            acquire_ok = self._websocket_lock.acquire(timeout=self._websocket_lock_timeout)
            if not acquire_ok:
                raise TimeoutError('Websocket lock acquire timeout')

            addr = ws.remote_address
            if addr in self._websocket_locks:
                self._websocket_locks[addr].release()
        except Exception as e:
            self.logger.warning('Unhandled exception while releasing websocket lock: {}'.format(str(e)))
        finally:
            self._websocket_lock.release()

    def notify_web_clients(self, event):
        """ Notify all the connected web clients (over websocket) of a new event """
        import websockets

        async def send_event(ws):
            try:
                self._acquire_websocket_lock(ws)
                await ws.send(str(event))
            except Exception as e:
                self.logger.warning('Error on websocket send_event: {}'.format(e))
            finally:
                self._release_websocket_lock(ws)

        loop = get_or_create_event_loop()
        wss = self.active_websockets.copy()

        for _ws in wss:
            try:
                loop.run_until_complete(send_event(_ws))
            except websockets.exceptions.ConnectionClosed:
                self.logger.warning('Websocket client {} connection lost'.format(_ws.remote_address))
                self.active_websockets.remove(_ws)
                if _ws.remote_address in self._websocket_locks:
                    del self._websocket_locks[_ws.remote_address]

    def websocket(self):
        """ Websocket main server """
        import websockets
        set_thread_name('WebsocketServer')

        async def register_websocket(websocket, path):
            address = websocket.remote_address if websocket.remote_address \
                else '<unknown client>'

            self.logger.info('New websocket connection from {} on path {}'.format(address, path))
            self.active_websockets.add(websocket)

            try:
                await websocket.recv()
            except websockets.exceptions.ConnectionClosed:
                self.logger.info('Websocket client {} closed connection'.format(address))
                self.active_websockets.remove(websocket)
                if address in self._websocket_locks:
                    del self._websocket_locks[address]

        websocket_args = {}
        if self.ssl_context:
            websocket_args['ssl'] = self.ssl_context

        loop = get_or_create_event_loop()
        loop.run_until_complete(
            websockets.serve(register_websocket, '0.0.0.0', self.websocket_port,
                             **websocket_args))
        loop.run_forever()

    def _start_web_server(self):
        def proc():
            self.logger.info('Starting local web server on port {}'.format(self.port))
            kwargs = {
                'host': '0.0.0.0',
                'port': self.port,
                'use_reloader': False,
                'debug': False,
            }

            if self.ssl_context:
                kwargs['ssl_context'] = self.ssl_context

            application.run(**kwargs)

        return proc

    def run(self):
        super().run()

        if not self.disable_websocket:
            self.logger.info('Initializing websocket interface')
            self.websocket_thread = threading.Thread(target=self.websocket)
            self.websocket_thread.start()

        if not self.run_externally:
            self.server_proc = Process(target=self._start_web_server(),
                                       name='WebServer')
            self.server_proc.start()
            self.server_proc.join()
        elif self.uwsgi_args:
            uwsgi_cmd = ['uwsgi'] + self.uwsgi_args
            self.logger.info('Starting uWSGI with arguments {}'.format(uwsgi_cmd))
            self.server_proc = subprocess.Popen(uwsgi_cmd)
        else:
            self.logger.info('The web server is configured to be launched externally but ' +
                             'no uwsgi_args were provided. Make sure that you run another external service' +
                             'for the webserver (e.g. nginx)')


# vim:sw=4:ts=4:et:



import re
import urllib.parse
import urllib.request

from platypush.context import get_plugin
# noinspection PyProtectedMember
from platypush.plugins.media.search import MediaSearcher


class YoutubeMediaSearcher(MediaSearcher):
    def search(self, query, **kwargs):
        """
        Performs a YouTube search either using the YouTube API (faster and
        recommended, it requires the :mod:`platypush.plugins.google.youtube`
        plugin to be configured) or parsing the HTML search results (fallback
        slower method)
        """

        self.logger.info('Searching YouTube for "{}"'.format(query))

        try:
            return self._youtube_search_api(query=query)
        except Exception as e:
            self.logger.warning('Unable to load the YouTube plugin, falling ' +
                                'back to HTML parse method: {}'.format(str(e)))

            return self._youtube_search_html_parse(query=query)

    @staticmethod
    def _youtube_search_api(query):
        return [
            {
                'url': 'https://www.youtube.com/watch?v=' + item['id']['videoId'],
                **item.get('snippet', {}),
            }
            for item in get_plugin('google.youtube').search(query=query).output
            if item.get('id', {}).get('kind') == 'youtube#video'
        ]

    def _youtube_search_html_parse(self, query):
        query = urllib.parse.quote(query)
        url = "https://www.youtube.com/results?search_query=" + query
        response = urllib.request.urlopen(url)
        html = response.read().decode('utf-8')
        results = []

        while html:
            m = re.search('(<a href="(/watch\?v=.+?)".+?yt-uix-tile-link.+?title="(.+?)".+?>)', html)
            if m:
                results.append({
                    'url': 'https://www.youtube.com' + m.group(2),
                    'title': m.group(3)
                })

                html = html.split(m.group(1))[1]
            else:
                html = ''

        self.logger.info('{} YouTube video results for the search query "{}"'
                         .format(len(results), query))

        return results


# vim:sw=4:ts=4:et:



import json
import logging
import threading

from redis import Redis

from platypush.bus import Bus
from platypush.config import Config
from platypush.message import Message

logger = logging.getLogger(__name__)


class RedisBus(Bus):
    """ Overrides the in-process in-memory local bus with a Redis bus """
    _DEFAULT_REDIS_QUEUE = 'platypush/bus'

    def __init__(self, on_message=None, redis_queue=_DEFAULT_REDIS_QUEUE,
                 *args, **kwargs):
        super().__init__(on_message=on_message)

        if not args and not kwargs:
            kwargs = (Config.get('backend.redis') or {}).get('redis_args', {})

        self.redis = Redis(*args, **kwargs)
        self.redis_args = kwargs
        self.redis_queue = redis_queue
        self.on_message = on_message
        self.thread_id = threading.get_ident()

    def get(self):
        """ Reads one message from the Redis queue """
        msg = None

        try:
            msg = self.redis.blpop(self.redis_queue)
            if msg and msg[1]:
                msg = Message.build(json.loads(msg[1].decode('utf-8')))
            else:
                msg = None
        except Exception as e:
            logger.exception(e)

        return msg

    def post(self, msg):
        """ Sends a message to the Redis queue """
        return self.redis.rpush(self.redis_queue, str(msg))


# vim:sw=4:ts=4:et:




from platypush.plugins import action
from platypush.plugins.google import GooglePlugin


class GoogleFitPlugin(GooglePlugin):
    """
    Google Fit plugin
    """

    scopes = ['https://www.googleapis.com/auth/fitness.activity.read',
              'https://www.googleapis.com/auth/fitness.body.read',
              'https://www.googleapis.com/auth/fitness.body_temperature.read',
              'https://www.googleapis.com/auth/fitness.location.read']

    def __init__(self, user_id='me', *args, **kwargs):
        """
        :param user_id: Default Google user_id (default: 'me', default
            configured account user)
        :type user_id: str or int
        """

        super().__init__(scopes=self.scopes, *args, **kwargs)
        self.user_id = user_id


    @action
    def get_data_sources(self, user_id=None):
        """
        Get the available data sources for the specified user_id
        """

        service = self.get_service(service='fitness', version='v1')
        sources = service.users().dataSources(). \
            list(userId=user_id or self.user_id).execute()

        return sources['dataSource']

    @action
    def get_data(self, data_source_id, user_id=None, limit=None):
        """
        Get raw data for the specified data_source_id

        :param data_source_id: Data source ID, see `get_data_sources`
        :type data_source_id: str
        """

        service = self.get_service(service='fitness', version='v1')
        kwargs = {
            'dataSourceId': data_source_id,
            'userId': user_id or self.user_id,
        }

        if limit:
            kwargs['limit'] = limit
        data_points = []

        for data_point in service.users().dataSources().dataPointChanges(). \
                list(**kwargs).execute().get('insertedDataPoint', []):
            data_point['startTime'] = float(data_point.pop('startTimeNanos'))/1e9
            data_point['endTime'] = float(data_point.pop('endTimeNanos'))/1e9
            data_point['modifiedTime'] = float(data_point.pop('modifiedTimeMillis'))/1e6
            values = []

            for value in data_point.pop('value'):
                if value.get('intVal') is not None:
                    value = value['intVal']
                elif value.get('fpVal') is not None:
                    value = value['fpVal']
                elif value.get('stringVal') is not None:
                    value = value['stringVal']
                elif value.get('mapVal'):
                    value = {
                        v['key']: v['value'].get(
                            'intVal', v['value'].get(
                                'fpVal', v['value'].get('stringVal')))
                        for v in value['mapVal'] }

                values.append(value)

            data_point['values'] = values
            data_points.append(data_point)

        return data_points


# vim:sw=4:ts=4:et:



import time

from threading import Timer
from multiprocessing import Process

import Leap

from platypush.backend import Backend
from platypush.context import get_backend
from platypush.message.event.sensor.leap import LeapFrameEvent, \
    LeapFrameStartEvent, LeapFrameStopEvent, LeapConnectEvent, LeapDisconnectEvent


class SensorLeapBackend(Backend):
    """
    Backend for events generated using a Leap Motion device to track hands and
    gestures, https://www.leapmotion.com/

    Note that the default SDK is not compatible with Python 3. Follow the
    instructions on https://github.com/BlackLight/leap-sdk-python3 to build the
    Python 3 module.

    Also, you'll need the Leap driver and utils installed on your OS (follow
    instructions at https://www.leapmotion.com/setup/) and the `leapd` daemon
    running to recognize your controller.

    Requires:

        * The Redis backend enabled
        * The Leap Motion SDK compiled with Python 3 support, see my port at https://github.com:BlackLight/leap-sdk-python3.git
        * The `leapd` daemon to be running and your Leap Motion connected

    Triggers:

        * :class:`platypush.message.event.sensor.leap.LeapFrameEvent` when a new frame is received
        * :class:`platypush.message.event.sensor.leap.LeapFrameStartEvent` when a new sequence of frame starts
        * :class:`platypush.message.event.sensor.leap.LeapFrameStopEvent` when a sequence of frame stops
        * :class:`platypush.message.event.sensor.leap.LeapConnectEvent` when a Leap Motion device is connected
        * :class:`platypush.message.event.sensor.leap.LeapDisconnectEvent` when a Leap Motion device disconnects
    """

    _listener_proc = None

    def __init__(self,
                 position_ranges=None,
                 position_tolerance=0.0,  # Position variation tolerance in %
                 frames_throttle_secs=None,
                 *args, **kwargs):
        """
        :param position_ranges: It specifies how wide the hand space (x, y and z axes) should be in millimiters.

        Default::

            [
                [-300.0, 300.0],  # x axis
                [25.0, 600.0],    # y axis
                [-300.0, 300.0],  # z axis
            ]

        :type position_ranges: list[list[float]]

        :param position_tolerance: % of change between a frame and the next to really consider the next frame as a new one (default: 0)
        :type position_tolerance: float

        :param frames_throttle_secs: If set, the frame events will be throttled
            and pushed to the main queue at the specified rate. Good to set if
            you want to connect Leap Motion events to actions that have a lower
            throughput (the Leap Motion can send a lot of frames per second).
            Default: None (no throttling)
        :type frames_throttle_secs: float
        """

        super().__init__(*args, **kwargs)

        if position_ranges is None:
            position_ranges = [
                [-300.0, 300.0],  # x axis
                [25.0, 600.0],  # y axis
                [-300.0, 300.0],  # z axis
            ]

        self.position_ranges = position_ranges
        self.position_tolerance = position_tolerance
        self.frames_throttle_secs = frames_throttle_secs

    def run(self):
        super().run()

        def _listener_process():
            listener = LeapListener(position_ranges=self.position_ranges,
                                    position_tolerance=self.position_tolerance,
                                    frames_throttle_secs=self.frames_throttle_secs,
                                    logger=self.logger)

            controller = Leap.Controller()

            if not controller:
                raise RuntimeError('No Leap Motion controller found - is your ' +
                                'device connected and is leapd running?')

            controller.add_listener(listener)
            self.logger.info('Leap Motion backend initialized')

            while not self.should_stop():
                time.sleep(0.1)

        time.sleep(1)
        self._listener_proc = Process(target=_listener_process)
        self._listener_proc.start()
        self._listener_proc.join()


class LeapFuture(Timer):
    def __init__(self, seconds, listener, event):
        self.listener = listener
        self.event = event

        super().__init__(seconds, self._callback_wrapper())

    def _callback_wrapper(self):
        def _callback():
            self.listener._send_event(self.event)
        return _callback


class LeapListener(Leap.Listener):
    def __init__(self, position_ranges, position_tolerance, logger,
                 frames_throttle_secs=None):
        super().__init__()

        self.prev_frame = None
        self.position_ranges = position_ranges
        self.position_tolerance = position_tolerance
        self.frames_throttle_secs = frames_throttle_secs
        self.logger = logger
        self.running_future = None

    def _send_event(self, event):
        backend = get_backend('redis')
        if not backend:
            self.logger.warning('Redis backend not configured, I cannot propagate the following event: {}'.
                                format(event))
            return

        backend.send_message(event)

    def send_event(self, event):
        if self.frames_throttle_secs:
            if not self.running_future or not self.running_future.is_alive():
                self.running_future = LeapFuture(seconds=self.frames_throttle_secs,
                                                 listener=self, event=event)
                self.running_future.start()
        else:
            self._send_event(event)

    def on_init(self, controller):
        self.prev_frame = None
        self.logger.info('Leap controller listener initialized')

    def on_connect(self, controller):
        self.logger.info('Leap controller connected')
        self.prev_frame = None
        self.send_event(LeapConnectEvent())

    def on_disconnect(self, controller):
        self.logger.info('Leap controller disconnected')
        self.prev_frame = None
        self.send_event(LeapDisconnectEvent())

    def on_exit(self, controller):
        self.logger.info('Leap listener terminated')

    def on_frame(self, controller):
        frame = controller.frame()

        if len(frame.hands) > 0:
            hands = self._flatten_hands(frame)
            if hands:
                if not self.prev_frame:
                    self.send_event(LeapFrameStartEvent())
                self.send_event(LeapFrameEvent(hands=hands))
            self.prev_frame = frame
        else:
            if self.prev_frame:
                self.send_event(LeapFrameStopEvent())
            self.prev_frame = None

    def _flatten_hands(self, frame):
        return [
            {
                'confidence': hand.confidence,
                'direction': [hand.direction[0], hand.direction[1], hand.direction[2]],
                'id': hand.id,
                'is_left': hand.is_left,
                'is_right': hand.is_right,
                'palm_normal': [hand.palm_normal[0], hand.palm_normal[1], hand.palm_normal[2]],
                'palm_position': self._normalize_position(hand.palm_position),
                'palm_velocity': [hand.palm_velocity[0], hand.palm_velocity[1], hand.palm_velocity[2]],
                'palm_width': hand.palm_width,
                'sphere_center': [hand.sphere_center[0], hand.sphere_center[1], hand.sphere_center[2]],
                'sphere_radius': hand.sphere_radius,
                'stabilized_palm_position': self._normalize_position(hand.stabilized_palm_position),
                'time_visible': hand.time_visible,
                'wrist_position': self._normalize_position(hand.wrist_position),
            }
            for i, hand in enumerate(frame.hands)
            if hand.is_valid and (
                len(frame.hands) != len(self.prev_frame.hands) or
                self._position_changed(
                    old_position=self.prev_frame.hands[i].stabilized_palm_position,
                    new_position=hand.stabilized_palm_position)

                if self.prev_frame
                else True
            )
        ]

    def _normalize_position(self, position):
        # Normalize absolute position onto a hemisphere centered in (0,0)
        # having x_range = z_range = [-100, 100], y_range = [0, 100]

        return [
            self._scale_scalar(value=position[0], range=self.position_ranges[0], new_range=[-100.0, 100.0]),
            self._scale_scalar(value=position[1], range=self.position_ranges[1], new_range=[0.0, 100.0]),
            self._scale_scalar(value=position[2], range=self.position_ranges[2], new_range=[-100.0, 100.0]),
        ]

    @staticmethod
    def _scale_scalar(value, range, new_range):
        if value < range[0]:
            value=range[0]
        if value > range[1]:
            value=range[1]

        return ((new_range[1]-new_range[0])/(range[1]-range[0]))*(value-range[0]) + new_range[0]

    def _position_changed(self, old_position, new_position):
        return (
            abs(old_position[0]-new_position[0]) > self.position_tolerance or
            abs(old_position[1]-new_position[1]) > self.position_tolerance or
            abs(old_position[2]-new_position[2]) > self.position_tolerance)


# vim:sw=4:ts=4:et:



import hashlib
import requests

from enum import Enum
from typing import Optional, Union, List

from platypush.message.response.pihole import PiholeStatusResponse
from platypush.plugins import Plugin, action


class PiholeStatus(Enum):
    ENABLED = 'enabled'
    DISABLED = 'disabled'


class PiholePlugin(Plugin):
    """
    Plugin for interacting with a `Pi-Hole <https://pi-hole.net>`_ DNS server for advertisement and content blocking.
    """

    def __init__(self, server: Optional[str] = None, password: Optional[str] = None, api_key: Optional[str] = None,
                 ssl: bool = False, verify_ssl: bool = True, **kwargs):
        """
        :param server: Default Pi-hole server IP address.
        :param password: Password for the default Pi-hole server.
        :param api_key: Alternatively to the password, you can also provide the server ``api_key``, as retrieved from
            http://pi-hole-server/admin/scripts/pi-hole/php/api_token.php
        :param ssl: Set to true if the host uses HTTPS (default: False).
        :param verify_ssl: Set to False to disable SSL certificate check.
        """
        super().__init__(**kwargs)
        self.server = server
        self.password = password
        self.api_key = api_key
        self.ssl = ssl
        self.verify_ssl = verify_ssl

    @staticmethod
    def _get_token(password: Optional[str] = None, api_key: Optional[str] = None) -> str:
        if not password:
            return api_key or ''
        return hashlib.sha256(hashlib.sha256(str(password).encode()).hexdigest().encode()).hexdigest()

    def _get_url(self, name: str, server: Optional[str] = None, password: Optional[str] = None,
                 ssl: Optional[bool] = None, api_key: Optional[str] = None, **kwargs) -> str:
        if not server:
            server = self.server
            password = password or self.password
            api_key = api_key or self.api_key
            ssl = ssl if ssl is not None else self.ssl

        args = '&'.join(['{key}={value}'.format(key=key, value=value) for key, value in kwargs.items()
                         if value is not None])

        if args:
            args = '&' + args

        token = self._get_token(password=password, api_key=api_key)
        if token:
            token = '&auth=' + token

        return 'http{ssl}://{server}/admin/api.php?{name}{token}{args}'.format(
            ssl='s' if ssl else '', server=server, name=name, token=token, args=args)

    @staticmethod
    def _normalize_number(n: Union[str, int]):
        return int(str(n).replace(',', ''))

    @action
    def status(self, server: Optional[str] = None, password: Optional[str] = None, api_key: Optional[str] = None,
               ssl: bool = None) \
            -> PiholeStatusResponse:
        """
        Get the status and statistics of a running Pi-hole server.

        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        :return: :class:`platypush.message.response.pihole.PiholeStatusResponse`
        """
        status = requests.get(self._get_url('summary', server=server, password=password, api_key=api_key,
                                            ssl=ssl), verify=self.verify_ssl).json()
        version = requests.get(self._get_url('versions', server=server, password=password, api_key=api_key,
                                             ssl=ssl), verify=self.verify_ssl).json()

        return PiholeStatusResponse(
            server=server or self.server,
            status=PiholeStatus(status.get('status')).value,
            ads_percentage=float(status.get('ads_percentage_today')),
            blocked=self._normalize_number(status.get('ads_blocked_today')),
            cached=self._normalize_number(status.get('queries_cached')),
            domain_count=self._normalize_number(status.get('domains_being_blocked')),
            forwarded=self._normalize_number(status.get('queries_forwarded')),
            queries=self._normalize_number(status.get('dns_queries_today')),
            total_clients=self._normalize_number(status.get('clients_ever_seen')),
            total_queries=self._normalize_number(status.get('dns_queries_all_types')),
            unique_clients=self._normalize_number(status.get('unique_clients')),
            unique_domains=self._normalize_number(status.get('unique_domains')),
            version=version.get('core_current'),
        )

    @action
    def enable(self, server: Optional[str] = None, password: Optional[str] = None, api_key: Optional[str] = None,
               ssl: bool = None):
        """
        Enable a Pi-hole server.

        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        response = requests.get(self._get_url('enable', server=server, password=password, api_key=api_key, ssl=ssl),
                                verify=self.verify_ssl)

        try:
            status = (response.json() or {}).get('status')
        except Exception as e:
            raise AssertionError('Could not enable the server: {}'.format(response.text or str(e)))

        assert status == 'enabled', 'Could not enable the server: Wrong credentials'
        return response.json()

    @action
    def disable(self, server: Optional[str] = None, password: Optional[str] = None, api_key: Optional[str] = None,
                seconds: Optional[int] = None, ssl: bool = None):
        """
        Disable a Pi-hole server.

        :param seconds: How long the server will be disabled in seconds (default: None, indefinitely).
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        if seconds:
            response = requests.get(self._get_url('', server=server, password=password, api_key=api_key,
                                                  ssl=ssl, disable=seconds), verify=self.verify_ssl)
        else:
            response = requests.get(self._get_url('disable', server=server, password=password, api_key=api_key,
                                                  ssl=ssl), verify=self.verify_ssl)

        try:
            status = (response.json() or {}).get('status')
        except Exception as e:
            raise AssertionError('Could not disable the server: {}'.format(response.text or str(e)))

        assert status == 'disabled', 'Could not disable the server: Wrong credentials'
        return response.json()

    def _list_manage(self, domain: str, list_name: str, endpoint: str, server: Optional[str] = None,
                     password: Optional[str] = None, api_key: Optional[str] = None, ssl: bool = None):
        data = {
            'list': list_name,
            'domain': domain
        }

        if password or self.password:
            data['pw'] = password or self.password
        elif api_key or self.api_key:
            data['auth'] = api_key or self.api_key

        base_url = "http{ssl}://{host}/admin/scripts/pi-hole/php/{endpoint}.php".format(
            ssl='s' if ssl or self.ssl else '', host=server or self.server, endpoint=endpoint
        )

        with requests.session() as s:
            s.get(base_url, verify=self.verify_ssl)
            response = requests.post(base_url, data=data, verify=self.verify_ssl).text.strip()

        return {'response': response}

    def _list_get(self, list_name: str, server: Optional[str] = None, ssl: bool = None) -> List[str]:
        response = requests.get("http{ssl}://{host}/admin/scripts/pi-hole/php/get.php?list={list}".format(
            ssl='s' if ssl or self.ssl else '', host=server or self.server, list=list_name
        ), verify=self.verify_ssl).json()

        ret = set()
        for ll in response:
            ret.update(ll)
        return list(ret)

    @action
    def get_blacklist(self, server: Optional[str] = None, ssl: bool = None) -> List[str]:
        """
        Get the content of the blacklist.

        :param server: Server IP address (default: default configured ``server`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_get(list_name='black', server=server, ssl=ssl)

    @action
    def get_whitelist(self, server: Optional[str] = None, ssl: bool = None) -> List[str]:
        """
        Get the content of the whitelist.

        :param server: Server IP address (default: default configured ``server`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_get(list_name='white', server=server, ssl=ssl)

    @action
    def get_list(self, list_name: str, server: Optional[str] = None, ssl: bool = None) -> List[str]:
        """
        Get the content of a list stored on the server.

        :param list_name: List name
        :param server: Server IP address (default: default configured ``server`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_get(list_name=list_name, server=server, ssl=ssl)

    @action
    def blacklist_add(self, domain: str, server: Optional[str] = None, password: Optional[str] = None,
                      api_key: Optional[str] = None, ssl: bool = None):
        """
        Add a domain to the blacklist.

        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(domain=domain, list_name='black', endpoint='add', server=server, password=password,
                                 api_key=api_key, ssl=ssl)

    @action
    def blacklist_remove(self, domain: str, server: Optional[str] = None, password: Optional[str] = None,
                         api_key: Optional[str] = None, ssl: bool = None):
        """
        Remove a domain from the blacklist.

        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(domain=domain, list_name='black', endpoint='sub', server=server, password=password,
                                 api_key=api_key, ssl=ssl)

    @action
    def whitelist_add(self, domain: str, server: Optional[str] = None, password: Optional[str] = None,
                      api_key: Optional[str] = None, ssl: bool = None):
        """
        Add a domain to the whitelist.

        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(domain=domain, list_name='white', endpoint='add', server=server, password=password,
                                 api_key=api_key, ssl=ssl)

    @action
    def whitelist_remove(self, domain: str, server: Optional[str] = None, password: Optional[str] = None,
                         api_key: Optional[str] = None, ssl: bool = None):
        """
        Remove a domain from the whitelist.

        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(domain=domain, list_name='white', endpoint='sub', server=server, password=password,
                                 api_key=api_key, ssl=ssl)

    @action
    def list_add(self, list_name: str,  domain: str, server: Optional[str] = None, password: Optional[str] = None,
                 api_key: Optional[str] = None, ssl: bool = None):
        """
        Add a domain to a custom list stored on the server.

        :param list_name: List name
        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(domain=domain, list_name=list_name, endpoint='add', server=server, password=password,
                                 api_key=api_key, ssl=ssl)

    @action
    def list_remove(self, list_name: str,  domain: str, server: Optional[str] = None, password: Optional[str] = None,
                    api_key: Optional[str] = None, ssl: bool = None):
        """
        Remove a domain from a custom list stored on the server.

        :param list_name: List name
        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(domain=domain, list_name=list_name, endpoint='sub', server=server, password=password,
                                 api_key=api_key, ssl=ssl)


# vim:sw=4:ts=4:et:



import os
import tempfile

from flask import Response, Blueprint, send_from_directory

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate, send_request
from platypush.config import Config
from platypush.plugins.camera.pi import CameraPiPlugin

camera_pi = Blueprint('camera.pi', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    camera_pi,
]


def video_feed():
    camera_conf = Config.get('camera.pi') or {}
    camera = CameraPiPlugin(**camera_conf)

    with camera:
        while True:
            output = camera.get_stream()

            with output.ready:
                output.ready.wait()
                frame = output.frame

            if frame and len(frame):
                yield (b'--frame\r\n'
                       b'Content-Type: image/jpeg\r\n\r\n' + frame + b'\r\n')


@camera_pi.route('/camera/pi/frame', methods=['GET'])
@authenticate()
def get_frame_img():
    filename = os.path.join(tempfile.gettempdir(), 'camera_pi.jpg')
    response = send_request('camera.pi.take_picture', image_file=filename)
    frame_file = (response.output or {}).get('image_file')
    assert frame_file is not None

    return send_from_directory(os.path.dirname(frame_file),
                               os.path.basename(frame_file))


@camera_pi.route('/camera/pi/stream', methods=['GET'])
@authenticate()
def get_stream_feed():
    return Response(video_feed(),
                    mimetype='multipart/x-mixed-replace; boundary=frame')


# vim:sw=4:ts=4:et:



import json
import time

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.todoist import NewItemEvent, RemovedItemEvent, ModifiedItemEvent, CheckedItemEvent, \
    ItemContentChangeEvent, TodoistSyncRequiredEvent

from platypush.plugins.todoist import TodoistPlugin


class TodoistBackend(Backend):
    """
    This backend listens for events on a remote Todoist account.

    Requires:

        * **todoist-python** (``pip install todoist-python``)
        * **websocket-client** (``pip install websocket-client``)

    Triggers:

        * :class:`platypush.message.event.todoist.NewItemEvent` when a new item is created.
        * :class:`platypush.message.event.todoist.RemovedItemEvent` when an item is removed.
        * :class:`platypush.message.event.todoist.CheckedItemEvent` when an item is checked.
        * :class:`platypush.message.event.todoist.ItemContentChangeEvent` when the content of an item is changed.
        * :class:`platypush.message.event.todoist.ModifiedItemEvent` when an item is changed and the change
            doesn't fall into the categories above.
        * :class:`platypush.message.event.todoist.TodoistSyncRequiredEvent` when an update has occurred that doesn't
            fall into the categories above and a sync is required to get up-to-date.

    """

    def __init__(self, api_token: str = None, **kwargs):
        super().__init__(**kwargs)
        self._plugin: TodoistPlugin = get_plugin('todoist')

        if not api_token:
            assert self._plugin and self._plugin.api_token, 'No api_token specified either on Todoist backend or plugin'
            self.api_token = self._plugin.api_token
        else:
            self.api_token = api_token

        self.url = self._plugin.get_user().output['websocket_url']
        self._ws = None
        self._connected = False
        self._todoist_initialized = False

        self._items = {}
        self._event_handled = False

    def _on_msg(self):
        # noinspection PyUnusedLocal
        def hndl(ws, msg):
            msg = json.loads(msg)
            if msg.get('type') == 'sync_needed':
                self._refresh_all()

        return hndl

    def _on_error(self):
        # noinspection PyUnusedLocal
        def hndl(ws, error):
            self.logger.warning('Todoist websocket error: {}'.format(error))
        return hndl

    def _on_close(self):
        # noinspection PyUnusedLocal
        def hndl(ws):
            self._ws = None
            self.logger.warning('Todoist websocket connection closed')
            self._connected = False

            while not self._connected:
                self._connect()
                time.sleep(10)

        return hndl

    def _on_open(self):
        # noinspection PyUnusedLocal
        def hndl(ws):
            self._connected = True
            self.logger.info('Todoist websocket connected')

            if not self._todoist_initialized:
                self._refresh_all()
                self._todoist_initialized = True

        return hndl

    def _connect(self):
        import websocket

        if not self._ws:
            self._ws = websocket.WebSocketApp(self.url,
                                              on_message=self._on_msg(),
                                              on_error=self._on_error(),
                                              on_close=self._on_close())

    def _refresh_items(self):
        new_items = {
            i['id']: i
            for i in self._plugin.get_items().output
        }

        if self._todoist_initialized:
            for id, item in new_items.items():
                if id not in self._items.keys():
                    self._event_handled = True
                    self.bus.post(NewItemEvent(item))

            for id, item in self._items.items():
                if id not in new_items.keys():
                    self._event_handled = True
                    self.bus.post(RemovedItemEvent(item))
                elif new_items[id] != item:
                    if new_items[id]['checked'] != item['checked']:
                        self._event_handled = True
                        self.bus.post(CheckedItemEvent(new_items[id]))
                    elif new_items[id]['is_deleted'] != item['is_deleted']:
                        self._event_handled = True
                        self.bus.post(RemovedItemEvent(new_items[id]))
                    elif new_items[id]['content'] != item['content']:
                        self._event_handled = True
                        self.bus.post(ItemContentChangeEvent(new_items[id]))
                    else:
                        self._event_handled = True
                        self.bus.post(ModifiedItemEvent(new_items[id]))

        self._items = new_items

    def _refresh_all(self):
        self._event_handled = False
        self._plugin.sync()
        self._refresh_items()

        if not self._event_handled:
            self.bus.post(TodoistSyncRequiredEvent())

    def run(self):
        super().run()
        self.logger.info('Started Todoist backend')

        self._connect()
        self._ws.on_open = self._on_open()
        self._ws.run_forever()


# vim:sw=4:ts=4:et:



import time

import platypush.plugins.gpio.zeroborg.lib as ZeroBorg

from platypush.backend import Backend
from platypush.message.event.sensor.ir import IrKeyUpEvent, IrKeyDownEvent


class SensorIrZeroborgBackend(Backend):
    """
    This backend will read for events on the infrared sensor of a ZeroBorg
    (https://www.piborg.org/motor-control-1135/zeroborg) circuitry for
    Raspberry Pi. You can see the codes associated to an IR event from any
    remote by running the scan utility::

        python -m platypush.backend.sensor.ir.zeroborg.scan

    Triggers:

        * :class:`platypush.message.event.sensor.ir.IrKeyDownEvent` when a key is pressed
        * :class:`platypush.message.event.sensor.ir.IrKeyUpEvent` when a key is released
    """

    last_message =None
    last_message_timestamp = None

    def __init__(self, no_message_timeout=0.37, **kwargs):
        super().__init__(**kwargs)
        self.no_message_timeout = no_message_timeout
        self.zb = ZeroBorg.ZeroBorg()
        self.zb.Init()
        self.logger.info('Initialized Zeroborg infrared sensor backend')


    def run(self):
        super().run()

        while True:
            try:
                self.zb.GetIrMessage()
                if self.zb.HasNewIrMessage():
                    message = self.zb.GetIrMessage()
                    if message != self.last_message:
                        self.logger.info('Received key down event on the IR sensor: {}'.format(message))
                        self.bus.post(IrKeyDownEvent(message=message))

                    self.last_message = message
                    self.last_message_timestamp = time.time()
            except OSError as e:
                self.logger.warning('Failed reading IR sensor status')

            if self.last_message_timestamp and \
                    time.time() - self.last_message_timestamp > self.no_message_timeout:
                self.logger.info('Received key up event on the IR sensor')
                self.bus.post(IrKeyUpEvent(message=self.last_message))

                self.last_message = None
                self.last_message_timestamp = None


# vim:sw=4:ts=4:et:




import ipaddress

from platypush.plugins import action
from platypush.plugins.switch import SwitchPlugin
from platypush.utils.workers import Workers
from .lib import WemoRunner
from .scanner import Scanner


class SwitchWemoPlugin(SwitchPlugin):
    """
    Plugin to control a Belkin WeMo smart switches
    (https://www.belkin.com/us/Products/home-automation/c/wemo-home-automation/)

    Requires:

        * **requests** (``pip install requests``)
    """

    _default_port = 49153

    def __init__(self, devices=None, netmask: str = None, port: int = _default_port, **kwargs):
        """
        :param devices: List of IP addresses or name->address map containing the WeMo Switch devices to control.
            This plugin previously used ouimeaux for auto-discovery but it's been dropped because
            1. too slow 2. too heavy 3. auto-discovery failed too often.
        :type devices: list or dict

        :param netmask: Alternatively to a list of static IP->name pairs,  you can specify the network mask where
            the devices should be scanned (e.g. '192.168.1.0/24')

        :param port: Port where the WeMo devices are expected to expose the RPC/XML over HTTP service (default: 49153)
        """

        super().__init__(**kwargs)
        self.port = port
        self.netmask = netmask
        self._devices = {}
        self._init_devices(devices)

    def _init_devices(self, devices):
        if devices:
            self._devices.update(devices if isinstance(devices, dict) else
                                 {addr: addr for addr in devices})
        else:
            self._devices = {}

        self._addresses = set(self._devices.values())

    @property
    def devices(self):
        """
        Get the list of available devices
        :returns: The list of devices.

        Example output::

            output = [
                    {
                        "ip": "192.168.1.123",
                        "name": "Switch 1",
                        "on": true,
                    },

                    {
                        # ...
                    }
                ]
        """

        return [
                self.status(device).output
                for device in self._devices.values()
        ]

    def _get_address(self, device: str) -> str:
        if device not in self._addresses:
            try:
                return self._devices[device]
            except KeyError:
                pass

        return device

    @action
    def status(self, device: str = None, *args, **kwargs):
        devices = {device: device} if device else self._devices.copy()

        ret = [
            {
                'id': addr,
                'ip': addr,
                'name': name if name != addr else WemoRunner.get_name(addr),
                'on': WemoRunner.get_state(addr),
            }
            for (name, addr) in devices.items()
        ]

        return ret[0] if device else ret

    @action
    def on(self, device: str, **kwargs):
        """
        Turn a switch on

        :param device: Device name or address
        """
        device = self._get_address(device)
        WemoRunner.on(device)
        return self.status(device)

    @action
    def off(self, device: str, **kwargs):
        """
        Turn a switch off

        :param device: Device name or address
        """
        device = self._get_address(device)
        WemoRunner.off(device)
        return self.status(device)

    @action
    def toggle(self, device: str, *args, **kwargs):
        """
        Toggle a device on/off state

        :param device: Device name or address
        """
        device = self._get_address(device)
        WemoRunner.toggle(device)
        return self.status(device)

    @action
    def get_state(self, device: str):
        """
        Get the on state of a device (True/False)

        :param device: Device name or address
        """
        device = self._get_address(device)
        return WemoRunner.get_state(device)

    @action
    def get_name(self, device: str):
        """
        Get the friendly name of a device

        :param device: Device name or address
        """
        device = self._get_address(device)
        return WemoRunner.get_name(device)

    @action
    def scan(self, netmask: str = None):
        netmask = netmask or self.netmask
        assert netmask, 'Scan not supported: No netmask specified'

        workers = Workers(10, Scanner, port=self.port)
        with workers:
            for addr in ipaddress.IPv4Network(netmask):
                workers.put(addr.exploded)

        devices = {
            dev.name: dev.addr
            for dev in workers.responses
        }

        self._init_devices(devices)
        return self.status()


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class ApplicationStartedEvent(Event):
    """
    Event triggered when the application has started and all the backends have been registered
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ApplicationStoppedEvent(Event):
    """
    Event triggered when the application stops
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# vim:sw=4:ts=4:et:




from platypush.message.response import Response


class PiholeStatusResponse(Response):
    def __init__(self,
                 server: str,
                 status: str,
                 ads_percentage: float,
                 blocked: int,
                 cached: int,
                 domain_count: int,
                 forwarded: int,
                 queries: int,
                 total_clients: int,
                 total_queries: int,
                 unique_clients: int,
                 unique_domains: int,
                 version: str,
                 *args,
                 **kwargs):
        super().__init__(*args, output={
            'server': server,
            'status': status,
            'ads_percentage': ads_percentage,
            'blocked': blocked,
            'cached': cached,
            'domain_count': domain_count,
            'forwarded': forwarded,
            'queries': queries,
            'total_clients': total_clients,
            'total_queries': total_queries,
            'unique_clients': unique_clients,
            'unique_domains': unique_domains,
            'version': version,
        }, **kwargs)


# vim:sw=4:ts=4:et:



import logging


class MediaSearcher:
    """
    Base class for media searchers
    """

    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger(self.__class__.__name__)

    def search(self, query, *args, **kwargs):
        raise NotImplementedError('The search method should be implemented ' +
                                  'by a derived class')


from .local import LocalMediaSearcher
from .youtube import YoutubeMediaSearcher
from .torrent import TorrentMediaSearcher

__all__ = ['LocalMediaSearcher', 'TorrentMediaSearcher', 'YoutubeMediaSearcher']


# vim:sw=4:ts=4:et:



import socket

from typing import Optional

from platypush.utils.workers import Worker
from .lib import WemoRunner


class ScanResult:
    def __init__(self, addr: str, name: str, on: bool):
        self.addr = addr
        self.name = name
        self.on = on


class Scanner(Worker):
    timeout = 1.5

    def __init__(self, port: int = WemoRunner.default_port, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.port = port

    def process(self, addr: str) -> Optional[ScanResult]:
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(self.timeout)
            sock.connect((addr, self.port))
            sock.close()

            return ScanResult(addr=addr, name=WemoRunner.get_name(addr), on=WemoRunner.get_state(addr))
        except OSError:
            pass


# vim:sw=4:ts=4:et:



from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.adafruit import ConnectedEvent, DisconnectedEvent, \
    FeedUpdateEvent


class AdafruitIoBackend(Backend):
    """
    Backend that listens to messages received over the Adafruit IO message queue

    Triggers:

        * :class:`platypush.message.event.adafruit.ConnectedEvent` when the
            backend connects to the Adafruit queue
        * :class:`platypush.message.event.adafruit.DisconnectedEvent` when the
            backend disconnects from the Adafruit queue
        * :class:`platypush.message.event.adafruit.FeedUpdateEvent` when an
            update event is received on a monitored feed

    Requires:

        * The :class:`platypush.plugins.adafruit.io.AdafruitIoPlugin` plugin to
            be active and configured.
    """

    def __init__(self, feeds, *args, **kwargs):
        """
        :param feeds: List of feed IDs to monitor
        :type feeds: list[str]
        """

        super().__init__(*args, **kwargs)
        self.feeds = feeds
        self._client = None

    def _init_client(self):
        if self._client:
            return

        from Adafruit_IO import MQTTClient
        plugin = get_plugin('adafruit.io')
        if not plugin:
            raise RuntimeError('Adafruit IO plugin not configured')

        self._client = MQTTClient(plugin._username, plugin._key)
        self._client.on_connect = self.on_connect()
        self._client.on_disconnect = self.on_disconnect()
        self._client.on_message = self.on_message()

    def on_connect(self):
        def _handler(client):
            for feed in self.feeds:
                client.subscribe(feed)
            self.bus.post(ConnectedEvent())
        return _handler

    def on_disconnect(self):
        def _handler(client):
            self.bus.post(DisconnectedEvent())
        return _handler

    def on_message(self):
        def _handler(client, feed, data):
            try: data = float(data)
            except: pass
            self.bus.post(FeedUpdateEvent(feed=feed, data=data))
        return _handler

    def run(self):
        super().run()

        self.logger.info(('Initialized Adafruit IO backend, listening on ' +
                          'feeds {}').format(self.feeds))

        while not self.should_stop():
            try:
                self._init_client()
                self._client.connect()
                self._client.loop_blocking()
            except Exception as e:
                self.logger.exception(e)
                self._client = None


# vim:sw=4:ts=4:et:



import enum
import urllib.parse

from platypush.context import get_bus
from platypush.plugins.media import MediaPlugin, PlayerState
from platypush.message.event.media import MediaPlayEvent, MediaPauseEvent, MediaStopEvent, MediaSeekEvent

from platypush.plugins import action


class PlayerEvent(enum.Enum):
    STOP = 'stop'
    PLAY = 'play'
    PAUSE = 'pause'


class MediaOmxplayerPlugin(MediaPlugin):
    """
    Plugin to control video and media playback using OMXPlayer.

    Requires:

        * **omxplayer** installed on your system (see your distro instructions)
        * **omxplayer-wrapper** (``pip install omxplayer-wrapper``)
    """

    def __init__(self, args=None, *argv, **kwargs):
        """
        :param args: Arguments that will be passed to the OMXPlayer constructor
            (e.g. subtitles, volume, start position, window size etc.) see
            https://github.com/popcornmix/omxplayer#synopsis and
            http://python-omxplayer-wrapper.readthedocs.io/en/latest/omxplayer/#omxplayer.player.OMXPlayer
        :type args: list
        """

        super().__init__(*argv, **kwargs)

        if args is None:
            args = []

        self.args = args
        self._player = None
        self._handlers = {e.value: [] for e in PlayerEvent}

    @action
    def play(self, resource, subtitles=None, *args, **kwargs):
        """
        Play a resource.

        :param resource: Resource to play. Supported types:

            * Local files (format: ``file://<path>/<file>``)
            * Remote videos (format: ``https://<url>/<resource>``)
            * YouTube videos (format: ``https://www.youtube.com/watch?v=<id>``)
            * Torrents (format: Magnet links, Torrent URLs or local Torrent files)

        :param subtitles: Subtitles file
        """

        if subtitles:
            args += ('--subtitles', subtitles)

        resource = self._get_resource(resource)
        if self._player:
            try:
                self._player.stop()
                self._player = None
            except Exception as e:
                self.logger.exception(e)
                self.logger.warning('Unable to stop a previously running instance ' +
                                    'of OMXPlayer, trying to play anyway')

        from dbus import DBusException

        try:
            from omxplayer import OMXPlayer
            self._player = OMXPlayer(resource, args=self.args)
            self._init_player_handlers()
        except DBusException as e:
            self.logger.warning('DBus connection failed: you will probably not ' +
                                'be able to control the media')
            self.logger.exception(e)

        self._player.pause()
        if self.volume:
            self.set_volume(volume=self.volume)

        self._player.play()
        return self.status()

    @action
    def pause(self):
        """ Pause the playback """
        if self._player:
            self._player.play_pause()
        return self.status()

    @action
    def stop(self):
        """ Stop the playback (same as quit) """
        return self.quit()

    @action
    def quit(self):
        """ Quit the player """
        from omxplayer.player import OMXPlayerDeadError

        if self._player:
            try:
                self._player.stop()
                self._player.quit()
            except OMXPlayerDeadError:
                pass

            self._player = None

        return {'status': 'stop'}

    @action
    def voldown(self):
        """ Volume down by 10% """
        if self._player:
            self._player.set_volume(max(0, self._player.volume()-0.1))
        return self.status()

    @action
    def volup(self):
        """ Volume up by 10% """
        if self._player:
            self._player.set_volume(min(1, self._player.volume()+0.1))
        return self.status()

    @action
    def back(self, offset=60):
        """ Back by (default: 60) seconds """
        if self._player:
            self._player.seek(-offset)
        return self.status()

    @action
    def forward(self, offset=60):
        """ Forward by (default: 60) seconds """
        if self._player:
            self._player.seek(offset)
        return self.status()

    @action
    def next(self):
        """ Play the next track/video """
        if self._player:
            self._player.stop()

        if self._videos_queue:
            video = self._videos_queue.pop(0)
            self.play(video)

        return self.status()

    @action
    def hide_subtitles(self):
        """ Hide the subtitles """
        if self._player:
            self._player.hide_subtitles()
        return self.status()

    @action
    def hide_video(self):
        """ Hide the video """
        if self._player:
            self._player.hide_video()
        return self.status()

    @action
    def is_playing(self):
        """
        :returns: True if it's playing, False otherwise
        """

        return self._player.is_playing()

    @action
    def load(self, resource, pause=False, **kwargs):
        """
        Load a resource/video in the player.

        :param resource: URL or filename to load
        :type resource: str

        :param pause: If set, load the video in paused mode (default: False)
        :type pause: bool
        """

        if self._player:
            self._player.load(resource, pause=pause)
        return self.status()

    @action
    def metadata(self):
        """ Get the metadata of the current video """
        if self._player:
            return self._player.metadata()
        return self.status()

    @action
    def mute(self):
        """ Mute the player """
        if self._player:
            self._player.mute()
        return self.status()

    @action
    def unmute(self):
        """ Unmute the player """
        if self._player:
            self._player.unmute()
        return self.status()

    @action
    def seek(self, relative_position):
        """
        Seek backward/forward by the specified number of seconds

        :param relative_position: Number of seconds relative to the current cursor
        :type relative_position: int
        """

        if self._player:
            self._player.seek(relative_position)
        return self.status()

    @action
    def set_position(self, position):
        """
        Seek backward/forward to the specified absolute position

        :param position: Number of seconds from the start
        :type position: int
        """

        if self._player:
            self._player.seek(position)
        return self.status()

    @action
    def set_volume(self, volume):
        """
        Set the volume

        :param volume: Volume value between 0 and 100
        :type volume: int
        """

        if self._player:
            self._player.set_volume(volume/100)
        return self.status()

    @action
    def status(self):
        """
        Get the current player state.

        :returns: A dictionary containing the current state.

        Format::

            output = {
                "duration": Duration in seconds,
                "filename": Media filename,
                "fullscreen": true or false,
                "mute": true or false,
                "path": Media path
                "pause": true or false,
                "position": Position in seconds
                "seekable": true or false
                "state": play, pause or stop
                "title": Media title
                "url": Media url
                "volume": Volume between 0 and 100
                "volume_max": 100,
            }
        """

        from omxplayer.player import OMXPlayerDeadError

        if not self._player:
            return {
                'state': PlayerState.STOP.value
            }

        try:
            state = self._player.playback_status().lower()
        except OMXPlayerDeadError:
            self._player = None
            return {
                'state': PlayerState.STOP.value
            }

        if state == 'playing':
            state = PlayerState.PLAY.value
        elif state == 'stopped':
            state = PlayerState.STOP.value
        elif state == 'paused':
            state = PlayerState.PAUSE.value

        return {
            'duration': self._player.duration(),
            'filename': urllib.parse.unquote(self._player.get_source()).split('/')[-1] if self._player.get_source().startswith('file://') else None,
            'fullscreen': self._player.fullscreen(),
            'mute': self._player._is_muted,
            'path': self._player.get_source(),
            'pause': state == PlayerState.PAUSE.value,
            'position': max(0, self._player.position()),
            'seekable': self._player.can_seek(),
            'state': state,
            'title': urllib.parse.unquote(self._player.get_source()).split('/')[-1] if self._player.get_source().startswith('file://') else None,
            'url': self._player.get_source(),
            'volume': self._player.volume() * 100,
            'volume_max': 100,
        }

    def add_handler(self, event_type, callback):
        if event_type not in self._handlers.keys():
            raise AttributeError('{} is not a valid PlayerEvent type'.
                                 format(event_type))

        self._handlers[event_type].append(callback)

    @staticmethod
    def _post_event(evt_type, **evt):
        bus = get_bus()
        bus.post(evt_type(player='local', plugin='media.omxplayer', **evt))

    def on_play(self):
        def _f(player):
            resource = player.get_source()
            self._post_event(MediaPlayEvent, resource=resource)
            for callback in self._handlers[PlayerEvent.PLAY.value]:
                callback(resource)

        return _f

    def on_pause(self):
        def _f(player):
            resource = player.get_source()
            self._post_event(MediaPauseEvent, resource=resource)
            for callback in self._handlers[PlayerEvent.PAUSE.value]:
                callback(resource)

        return _f

    def on_stop(self):
        def _f(player):
            self._post_event(MediaStopEvent)
            for callback in self._handlers[PlayerEvent.STOP.value]:
                callback()
        return _f

    def on_seek(self):
        def _f(player):
            self._post_event(MediaSeekEvent, position=player.position())
        return _f

    def _init_player_handlers(self):
        if not self._player:
            return

        self._player.playEvent += self.on_play()
        self._player.pauseEvent += self.on_pause()
        self._player.stopEvent += self.on_stop()
        self._player.positionEvent += self.on_seek()


# vim:sw=4:ts=4:et:



import json

from flask import Blueprint, abort, request, Response

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate, logger, send_message

from platypush.message.event.http.hook import WebhookEvent


hook = Blueprint('hook', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    hook,
]


@hook.route('/hook/<hook_name>', methods=['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'OPTIONS'])
@authenticate(skip_auth_methods=['session'])
def hook(hook_name):
    """ Endpoint for custom webhooks """

    event_args = {
        'hook': hook_name,
        'method': request.method,
        'args': dict(request.args or {}),
        'data': request.data.decode(),
    }

    if event_args['data']:
        try:
            event_args['data'] = json.loads(event_args['data'])
        except:
            pass

    event = WebhookEvent(**event_args)

    try:
        response = send_message(event)
        return Response(json.dumps({'status': 'ok', **event_args}),
                        mimetype='application/json')
    except Exception as e:
        logger().error('Error while dispatching webhook event {}: {}'.
                       format(event, str(e)))
        abort(500, str(e))


# vim:sw=4:ts=4:et:
